import logging
import os
import shutil

from twisted.internet.defer import inlineCallbacks, returnValue

from ensemble.errors import EnsembleError
from ensemble.formula.bundle import FormulaBundle
from ensemble.machine.unit import ServiceUnitDeployment
from ensemble.state.environment import EnvironmentStateManager
from ensemble.state.formula import FormulaStateManager
from ensemble.state.machine import MachineStateManager
from ensemble.state.service import ServiceStateManager

from .base import BaseAgent


log = logging.getLogger("ensemble.agents.machine")


class MachineAgent(BaseAgent):
    """An Ensemble Machine Agent.

    The machine agent is responsible for monitoring service units
    assigned to a machine. If a new unit is assigned to machine, the
    machine agent will download the formula, create a working
    space for the service unit agent, and then launch it.

    Additionally the machine agent will monitor the running service
    unit agents on the machine, via their ephemeral nodes, and
    restart them if they die.
    """

    name = "ensemble-machine-agent"
    unit_agent_module = "ensemble.agents.unit"

    @property
    def formulas_directory(self):
        return os.path.join(self.config["ensemble_directory"], "formulas")

    @property
    def units_directory(self):
        return os.path.join(self.config["ensemble_directory"], "units")

    @property
    def unit_state_directory(self):
        return os.path.join(self.config["ensemble_directory"], "state")

    @inlineCallbacks
    def start(self):
        """Start the machine agent.

        Creates state directories on the machine, retrieves the machine state,
        and enables watch on assigned units.
        """
        # initialize directory paths
        if not os.path.exists(self.formulas_directory):
            os.makedirs(self.formulas_directory)

        if not os.path.exists(self.units_directory):
            os.makedirs(self.units_directory)

        if not os.path.exists(self.unit_state_directory):
            os.makedirs(self.unit_state_directory)

        # get state managers we'll be utilizing
        self.service_state_manager = ServiceStateManager(self.client)
        self.formula_state_manager = FormulaStateManager(self.client)

        # retrieve the machine state for the machine we represent
        machine_manager = MachineStateManager(self.client)
        self.machine_state = yield machine_manager.get_machine_state(
            self.get_machine_id())

        # watch assigned units for the machine
        if self.get_watch_enabled():
            self.machine_state.watch_assigned_units(
                self.watch_service_units)

        # connect the machine agent
        yield self.machine_state.connect_agent()

    @inlineCallbacks
    def download_formula(self, formula_state):
        """Retrieve a formula from the provider storage to the local machine.

        Utilizes a local formula cache to avoid repeated downloading of the
        same formula.
        """
        # XXX this is a temporary hack, in future we should use signed urls
        # directly to the formula and bypass the provider file storage.
        environment_state_manager = EnvironmentStateManager(self.client)
        config = yield environment_state_manager.get_config()
        provider = config.get_default().get_machine_provider()
        file_storage = provider.get_file_storage()

        checksum = yield formula_state.get_sha256()
        formula_key = "%s:%s" % (formula_state.id, checksum)
        local_formula_path = os.path.join(
            self.formulas_directory, formula_key)

        remote_formula_path = "formulas/%s" % formula_key

        file_handle = yield file_storage.get(remote_formula_path)
        shutil.copyfileobj(
            file_handle, open(local_formula_path, "w"))

        returnValue(FormulaBundle(local_formula_path))

    @inlineCallbacks
    def watch_service_units(self, old_units, new_units):
        """Callback invoked when the assigned service units change.
        """
        if old_units is None:
            old_units = set()

        stopped = old_units - new_units
        started = new_units - old_units

        for unit_name in stopped:
            try:
                yield self.kill_service_unit(unit_name)
            except Exception, e:
                log.exception("Error stopping unit: %r", unit_name)

        for unit_name in started:
            try:
                yield self.start_service_unit(unit_name)
            except Exception, e:
                log.exception("Error starting unit: %r", unit_name)

    @inlineCallbacks
    def start_service_unit(self, service_unit_name):
        """Start a service unit on the machine.

        Downloads the formula, and extract it to the service unit directory,
        and launch the service unit agent within the unit directory.
        """
        # Retrieve the formula state to get at the formula.
        service_name, junk = service_unit_name.split("/")
        service_state = yield self.service_state_manager.get_service_state(
            service_name)
        formula_id = yield service_state.get_formula_id()
        formula_state = yield self.formula_state_manager.get_formula_state(
            formula_id)

        # Download the formula.
        bundle = yield self.download_formula(formula_state)

        # Use deployment to setup the workspace and start the unit agent.
        deployment = ServiceUnitDeployment(
            service_unit_name, self.units_directory)

        if not deployment.is_running():
            deployment.unpack_formula(bundle)
            yield deployment.start(self.get_machine_id(), self.client.servers)

    def kill_service_unit(self, service_unit_name):
        """Stop service unit and destroy disk state, ala SIGKILL or lxc-destroy
        """
        deployment = ServiceUnitDeployment(
            service_unit_name, self.units_directory)
        deployment.destroy()

    def get_machine_id(self):
        """Get the id of the machine as known within the zk state."""
        return self.config["machine_id"]

    def get_agent_name(self):
        return "machine:%s" % (self.get_machine_id())

    def configure(self, options):
        super(MachineAgent, self).configure(options)
        if not options.get("machine_id"):
            msg = ("--machine-id must be provided in the command line, "
                   "or $ENSEMBLE_MACHINE_ID in the environment")
            raise EnsembleError(msg)

    @classmethod
    def setup_options(cls, parser):
        super(MachineAgent, cls).setup_options(parser)

        machine_id = os.environ.get("ENSEMBLE_MACHINE_ID", "")
        parser.add_argument(
            "--machine-id", default=machine_id)
        return parser

if __name__ == '__main__':
    MachineAgent().run()
