import argparse
import yaml

from twisted.internet.defer import inlineCallbacks, returnValue

from ensemble.agents.unit import UnitAgent
from ensemble.agents.base import TwistedOptionNamespace
from ensemble.errors import EnsembleError

from ensemble.agents.tests.common import AgentTestBase
from ensemble.tests.common import get_test_zookeeper_address
from ensemble.unit.tests.test_workflow import WorkflowTestBase
from ensemble.state.errors import ServiceStateNotFound

from ensemble.hooks.tests.test_invoker import get_cli_environ_path


class UnitAgentTest(AgentTestBase, WorkflowTestBase):

    agent_class = UnitAgent

    @inlineCallbacks
    def setUp(self):
        yield super(UnitAgentTest, self).setUp()
        self.change_environment(
            PATH=get_cli_environ_path(),
            ENSEMBLE_UNIT_NAME="mysql/0")

    @inlineCallbacks
    def tearDown(self):
        yield self.agent.stop()
        yield super(UnitAgentTest, self).tearDown()

    @inlineCallbacks
    def get_agent_config(self):
        yield self.setup_default_test_relation()
        options = yield super(UnitAgentTest, self).get_agent_config()
        options["unit_name"] = str(self.states["unit"].unit_name)
        returnValue(options)

    def write_empty_hooks(self, start=True, stop=True, install=True, **kw):
        # NB Tests that use this helper method must properly wait on
        # the agent being stopped (yield self.agent.stopService()) to
        # avoid the environment being restored while asynchronously
        # the stop hook continues to execute. Otherwise
        # ENSEMBLE_UNIT_NAME, which hook invocation depends on, will
        # not be available and the stop hook will fail (somewhat
        # mysteriously!). The alternative is to set stop=False so that
        # the stop hook will not be created when writing the empty
        # hooks.
        output_file = self.makeFile()

        if install:
            self.write_hook(
                "install", "#!/bin/bash\necho install >> %s" % output_file)
        if start:
            self.write_hook(
                "start", "#!/bin/bash\necho start >> %s" % output_file)
        if stop:
            self.write_hook(
                "stop", "#!/bin/bash\necho stop >> %s" % output_file)

        for k in kw.keys():
            self.write_hook(k.replace("_", "-"),
                            "#!/bin/bash\necho $0 >> %s" % output_file)

        return output_file

    def parse_output(self, output_file):
        return filter(None, open(output_file).read().split("\n"))

    @inlineCallbacks
    def test_agent_start_stop_service(self):
        """Verify workflow state when starting and stopping the unit agent."""
        self.write_empty_hooks()
        yield self.agent.startService()
        current_state = yield self.agent.workflow.get_state()
        self.assertEqual(current_state, "started")
        yield self.agent.stopService()
        current_state = yield self.agent.workflow.get_state()
        self.assertEqual(current_state, "stopped")

    def test_agent_unit_name_environment_extraction(self):
        """Verify extraction of unit name from the environment."""
        self.change_args("unit-agent")
        self.change_environment(ENSEMBLE_UNIT_NAME="rabbit/1")
        parser = argparse.ArgumentParser()
        self.agent.setup_options(parser)
        options = parser.parse_args(namespace=TwistedOptionNamespace())
        self.assertEqual(options["unit_name"], "rabbit/1")

    def test_agent_unit_name_cli_extraction_error(self):
        """Failure to extract the unit name, results in a nice error message.
        """
        # We don't want ENSEMBLE_UNIT_NAME set, so that the expected
        # EnsembleError will be raised
        self.change_environment(
            PATH=get_cli_environ_path())
        self.change_args(
            "unit-agent",
            "--ensemble-directory", self.makeDir(),
            "--zookeeper-servers", get_test_zookeeper_address())

        parser = argparse.ArgumentParser()
        self.agent.setup_options(parser)
        options = parser.parse_args(namespace=TwistedOptionNamespace())

        e = self.assertRaises(EnsembleError,
                              self.agent.configure,
                              options)
        self.assertIn(
            ("--unit-name must be provided in the command line,"
            " or $ENSEMBLE_UNIT_NAME in the environment"),
            str(e))

    def test_agent_unit_name_cli_extraction(self):
        """The unit agent can parse its unit-name from the cli.
        """
        self.change_args("unit-agent", "--unit-name", "rabbit/1")
        parser = argparse.ArgumentParser()
        self.agent.setup_options(parser)
        options = parser.parse_args(namespace=TwistedOptionNamespace())
        self.assertEqual(options["unit_name"], "rabbit/1")

    def test_get_agent_name(self):
        self.assertEqual(self.agent.get_agent_name(), "unit:mysql/0")

    def test_agent_invalid_unit_name(self):
        """If the unit agent is given an invalid unit name, an error
        message is raised."""
        options = {}
        options["ensemble_directory"] = self.ensemble_directory
        options["zookeeper_servers"] = get_test_zookeeper_address()
        options["unit_name"] = "rabbit-1"
        agent = self.agent_class()
        agent.configure(options)
        return self.assertFailure(agent.startService(), ServiceStateNotFound)

    @inlineCallbacks
    def test_agent_agent_executes_install_and_start_hooks_on_startup(self):
        """On initial startup, the unit agent executes install and start hooks.
        """
        output_file = self.write_empty_hooks()
        hooks_complete = self.wait_on_hook(
            sequence=["install", "start"],
            executor=self.agent.executor)
        yield self.agent.startService()
        # Verify the hook has executed.
        yield hooks_complete
        self.assertEqual(self.parse_output(output_file),
                         ["install", "start"])
        yield self.assertState(self.agent.workflow, "started")
        yield self.agent.stopService()

    @inlineCallbacks
    def test_agent_executes_stop_hook_on_shutdown(self):
        """On shutdown, the agent stops the hook."""
        output_file = self.write_empty_hooks()
        yield self.agent.startService()
        hooks_complete = self.wait_on_hook(
            "stop", executor=self.agent.executor)
        yield self.agent.stopService()
        # Verify the hook has executed.
        yield hooks_complete
        self.assertEqual(self.parse_output(output_file),
                         ["install", "start", "stop"])
        yield self.assertState(self.agent.workflow, "stopped")

        # verify workflow state.
        f_state, history, zk_state = yield self.read_persistent_state(
            workflow=self.agent.workflow)
        self.assertEqual(f_state, zk_state)
        self.assertEqual(f_state, {"state": "stopped", "state_variables": {}})

    @inlineCallbacks
    def test_agent_executes_relation_changed_hook(self):
        """If a relation changes after the unit is started, a relation change
        hook is executed."""
        self.write_empty_hooks()
        file_path = self.makeFile()
        self.write_hook("app-relation-changed",
                ("#!/bin/sh\n"
                 "echo $ENSEMBLE_REMOTE_UNIT >> %s\n" % file_path))
        yield self.agent.startService()
        try:
            hook_complete = self.wait_on_hook(
                "app-relation-changed", executor=self.agent.executor)
            wordpress_states = yield self.add_opposite_service_unit(self.states)

            # Verify the hook has executed.
            yield hook_complete
            self.assertEqual(open(file_path).read().strip(),
                             wordpress_states["unit"].unit_name)
        finally:
            yield self.agent.stopService()

    @inlineCallbacks
    def test_agent_hook_api_usage(self):
        """If a relation changes after the unit is started, a relation change
        hook is executed."""
        self.write_empty_hooks()
        file_path = self.makeFile()
        self.write_hook("app-relation-changed",
                        "\n".join(
                            ["#!/bin/sh",
                             "echo `relation-list` >> %s" % file_path,
                             "echo `relation-set greeting=hello`",
                             "echo `relation-set planet=earth`",
                             "echo `relation-get %s planet` >> %s" % (
                                 self.states["unit"].unit_name, file_path)]))

        yield self.agent.startService()
        try:
            hook_complete = self.wait_on_hook(
                "app-relation-changed", executor=self.agent.executor)
            yield self.add_opposite_service_unit(self.states)

            # Verify the hook has executed.
            yield hook_complete

            # Verify hook output
            output = open(file_path).read().strip().split("\n")
            self.assertEqual(output, ["wordpress/0", "earth"])

            # Verify zookeeper state
            contents = yield self.states["unit_relation"].get_data()
            self.assertEqual({"greeting": "hello", "planet": "earth"},
                             yaml.load(contents))

            self.failUnlessIn("wordpress/0", output)
        finally:
            yield self.agent.stopService()

    @inlineCallbacks
    def test_agent_executes_depart_hook(self):
        """If a relation changes after the unit is started, a relation change
        hook is executed."""
        self.write_empty_hooks(app_relation_changed=True)
        file_path = self.makeFile()
        self.write_hook("app-relation-broken",
                ("#!/bin/sh\n"
                 "echo $ENSEMBLE_CHANGE >> %s\n" % file_path))
        yield self.agent.startService()
        try:
            hook_complete = self.wait_on_hook(
                "app-relation-changed", executor=self.agent.executor)

            yield self.add_opposite_service_unit(self.states)
            yield hook_complete

            # Watch the unit relation workflow complete
            workflow_complete = self.wait_on_state(
                self.agent.lifecycle.get_relation_workflow(
                    self.states["relation"].internal_id),
                "departed")

            yield self.relation_manager.remove_relation_state(
                self.states["relation"])

            hook_complete = self.wait_on_hook(
                "app-relation-broken", executor=self.agent.executor)

            # Verify the hook has executed.
            yield hook_complete
            self.assertEqual(open(file_path).read().strip(), "departed")

            # Wait for the workflow transition to complete.
            yield workflow_complete
            # Wait for the agent to properly stop before
            # tearDown restores the env
        finally:
            yield self.agent.stopService()

    @inlineCallbacks
    def test_agent_debug_watch(self):
        """The unit agent subscribes to changes to the hook debug settings.
        """
        self.agent.set_watch_enabled(True)
        yield self.agent.startService()

        yield self.states["unit"].enable_hook_debug(["*"])
        yield self.sleep(0.1)

        # Check the propogation to the executor
        self.assertNotEquals(
            self.agent.executor.get_hook_path("x"), "x")
