import os
import logging
from twisted.internet.defer import inlineCallbacks

from ensemble.errors import EnsembleError
from ensemble.state.service import ServiceStateManager
from ensemble.hooks.protocol import UnitSettingsFactory
from ensemble.hooks.executor import HookExecutor
from ensemble.unit.lifecycle import (
    UnitLifecycle, HOOK_SOCKET_FILE)
from ensemble.unit.workflow import UnitWorkflowState, UnitWorkflow

from ensemble.agents.base import BaseAgent


class UnitAgent(BaseAgent):
    """An Ensemble Unit Agent.

    Provides for the management of a formula, via hook execution in response to
    external events in the coordination space (zookeeper).
    """
    name = "ensemble-unit-agent"

    @classmethod
    def setup_options(cls, parser):
        super(UnitAgent, cls).setup_options(parser)
        unit_name = os.environ.get("ENSEMBLE_UNIT_NAME", "")
        parser.add_argument("--unit-name", default=unit_name)

    @property
    def unit_name(self):
        return self.config["unit_name"]

    def get_agent_name(self):
        return "unit:%s" % self.unit_name

    def configure(self, options):
        """Configure the unit agent."""
        super(UnitAgent, self).configure(options)
        if not options.get("unit_name"):
            msg = ("--unit-name must be provided in the command line, "
                   "or $ENSEMBLE_UNIT_NAME in the environment")
            raise EnsembleError(msg)
        self.executor = HookExecutor()
        self.api_factory = UnitSettingsFactory(
            self.executor.get_hook_context, logging.getLogger("unit.hook.api"))
        self.api_socket = None
        self.workflow = None

    @inlineCallbacks
    def start(self):
        """Start the unit agent process."""
        self.service_state_manager = ServiceStateManager(self.client)
        self.executor.start()

        # Retrieve our unit and configure working directories.
        service_name = self.unit_name.split("/")[0]
        service_state = yield self.service_state_manager.get_service_state(
            service_name)

        self.unit_state = yield service_state.get_unit_state(
            self.unit_name)
        self.unit_directory = os.path.join(
            self.config["ensemble_directory"],
            "units",
            self.unit_state.unit_name.replace("/", "-"))

        self.state_directory = os.path.join(
            self.config["ensemble_directory"], "state")

        # Setup the server portion of the cli api exposed to hooks.
        from twisted.internet import reactor
        self.api_socket = reactor.listenUNIX(
            os.path.join(self.unit_directory, HOOK_SOCKET_FILE),
            self.api_factory)

        # Inform the system, we're alive.
        yield self.unit_state.connect_agent()

        if self.get_watch_enabled():
            yield self.unit_state.watch_hook_debug(self.watch_hook_debug)

        self.lifecycle = UnitLifecycle(
            self.client,
            self.unit_state,
            service_state,
            self.unit_directory,
            self.executor)

        self.workflow = UnitWorkflowState(
            self.client,
            self.unit_state,
            self.lifecycle,
            UnitWorkflow,
            self.state_directory)

        # Fire initial transitions.
        yield self.workflow.transition_state("installed")
        yield self.workflow.transition_state("started")

    @inlineCallbacks
    def stop(self):
        """Stop the unit agent process."""
        if self.workflow:
            yield self.workflow.transition_state("stopped")
        if self.api_socket:
            yield self.api_socket.stopListening()
        yield self.api_factory.stopFactory()

    @inlineCallbacks
    def watch_hook_debug(self, change):
        """Update the hooks to be debugged when the settings change.
        """
        debug = yield self.unit_state.get_hook_debug()
        debug_hooks = debug and debug.get("debug_hooks") or None
        self.executor.set_debug(debug_hooks)

if __name__ == '__main__':
    UnitAgent.run()
