"""
Command for debugging hooks on a service unit.
"""
import os

from twisted.internet.defer import inlineCallbacks, returnValue

from ensemble.control.ssh import get_ip_address_for_unit
from ensemble.control.utils import get_environment

from ensemble.formula.errors import InvalidFormulaHook
from ensemble.state.formula import FormulaStateManager
from ensemble.state.service import ServiceStateManager


def configure_subparser(subparsers):
    sub_parser = subparsers.add_parser("debug-hooks", help=command.__doc__)
    sub_parser.add_argument(
        "-e", "--environment",
        help="Ensemble environment to operate in.")

    sub_parser.add_argument(
        "unit_name",
        help="Name of unit")

    sub_parser.add_argument(
        "hook_names", default=["*"], nargs="*",
        help="Name of hook, defaults to all")

    return sub_parser


@inlineCallbacks
def validate_hooks(client, unit_state, hook_names):

    # Assemble a list of valid hooks for the formula.
    valid_hooks = []
    service_manager = ServiceStateManager(client)
    endpoints = yield service_manager.get_relation_endpoints(
        unit_state.service_name)
    endpoint_names = [ep.relation_name for ep in endpoints]
    for endpoint_name in endpoint_names:
        valid_hooks.append("%s-relation-changed" % endpoint_name)
        valid_hooks.append("%s-relation-broken" % endpoint_name)
    valid_hooks.extend(["start", "stop", "install"])

    # Verify the debug names.
    for hook_name in hook_names:
        if hook_name in valid_hooks:
            continue
        break
    else:
        returnValue(True)

    # We dereference to the formula to give a fully qualified error
    # message.  I wish this was a little easier to dereference, the
    # service_manager.get_relation_endpoints effectively does this
    # already.
    service_manager = ServiceStateManager(client)
    service_state = yield service_manager.get_service_state(
        unit_state.service_name)
    formula_id = yield service_state.get_formula_id()
    formula_manager = FormulaStateManager(client)
    formula = yield formula_manager.get_formula_state(formula_id)
    raise InvalidFormulaHook(formula.id, hook_name)


@inlineCallbacks
def command(options):
    """Interactively debug a hook remotely on a service unit with screen.
    """
    environment = get_environment(options)
    provider = environment.get_machine_provider()
    client = yield provider.connect()

    options.log.debug("Retrieving unit and machine information.")
    ip_address, unit, machine = yield get_ip_address_for_unit(
        client, provider, options.unit_name)

    # Verify hook name
    if options.hook_names != ["*"]:
        options.log.debug("Verifying hook names...")
        yield validate_hooks(client, unit, options.hook_names)

    options.log.debug(
        "Enabling hook debug on unit (%r)..." % options.unit_name)
    yield unit.enable_hook_debug(options.hook_names)

    # Connect via ssh and start screen.
    options.log.info("Connecting to remote machine...")
    yield os.system("ssh -t ubuntu@%s screen -S %s-hook-debug" % (
        ip_address, options.unit_name.replace("/", "-")))

    options.log.info("Debug session ended.")
    # Ends hook debugging.
    yield client.close()
