""" Hook Execution.
"""
import os
import fnmatch
import logging
import tempfile

from twisted.internet.defer import (
    inlineCallbacks, DeferredQueue, Deferred, DeferredLock)


DEBUG_HOOK_TEMPLATE = """\
#!/bin/bash
ENSEMBLE_HOOK_NAME="%s"
ENSEMBLE_DEBUG=$(mktemp -d)
SESSION_NAME=$(echo "$ENSEMBLE_UNIT_NAME" | sed s/\\//-/g)
unset LS_COLORS
unset LESSOPEN
unset LESSCLOSE

env > $ENSEMBLE_DEBUG/env.sh
echo running > $ENSEMBLE_DEBUG/state

# Create internal script which will load the hook environment,
# and change the state file once the user is done with it.
cat > $ENSEMBLE_DEBUG/hook.sh <<END
#!/bin/sh
. $ENSEMBLE_DEBUG/env.sh
/bin/bash
echo stopped > $ENSEMBLE_DEBUG/state
END

chmod +x $ENSEMBLE_DEBUG/hook.sh

# Verify we have an open screen session.
if ! screen -ls | grep "\<$SESSION_NAME\>" > /dev/null;
then
  echo "Debug hook session not found."
  exit 0;
fi

# Create a new window in the existing session.
screen -S $SESSION_NAME -X screen -t $ENSEMBLE_HOOK_NAME
screen -S $SESSION_NAME -p $ENSEMBLE_HOOK_NAME -X exec $ENSEMBLE_DEBUG/hook.sh

while [ "$(cat $ENSEMBLE_DEBUG/state)" != stopped ]; do
    # Wait for user to quit the shell.
    sleep 1
fi

screen -S $SESSION_NAME -p $ENSEMBLE_HOOK_NAME -X kill
"""


class HookExecutor(object):
    """Executes scheduled hooks.

    A typical unit agent is subscribed to multiple event streams
    across unit and relation lifecycles. All of which will attempt to
    execute hooks in response to events. In order to serialize hook
    execution and bring observability, a hook executor is utilized
    across the different components that want to execute hooks.
    """

    STOP = object()

    def __init__(self):
        self._running = False
        self._executions = DeferredQueue()
        self._observer = None
        self._log = logging.getLogger("hook.executor")
        self._run_lock = DeferredLock()

        # The currently executing hook's context. None if no hook is executing.
        self._hook_context = None
        # The current names of hooks that should be debugged.
        self._debug_hook_names = None
        # The last utilized tempfile debug hook.
        self._debug_hook_file = None

    @inlineCallbacks
    def start(self):
        """Start the hook executor.

        After the executor is started, it will continue to serially execute
        any queued hook executions.
        """
        assert self._running is False, "Already running"
        self._running = True

        while self._running:

            next = yield self._executions.get()

            # The stop logic here is to allow for two different
            # scenarios. One is if the executor is currently waiting on
            # the queue, putting a stop value there will, immediately
            # wake it up and cause it to stop.

            # The other distinction is more subtle, if we invoke
            # start/stop/start on the executioner, and it was
            # currently executing a slow hook, then when the
            # executioner finishes with the hook it may now be in the
            # running state, resulting in two executioners closures
            # executing hooks. We track stops to ensure that only one
            # executioner closure is running at a time.

            if next is self.STOP:
                continue

            yield self._run_lock.acquire()
            if not self._running:
                self._run_lock.release()
                continue

            invoker, path, exec_deferred = next
            self._hook_context = invoker.get_context()
            hook_path = self.get_hook_path(path)

            self._log.debug("Running hook: %s", path)
            try:
                yield invoker(hook_path)
            except Exception, e:
                self._hook_context = None
                self._run_lock.release()
                self._log.debug("Hook error: %s %s", path, e)
                exec_deferred.errback(e)
            else:
                self._hook_context = None
                self._run_lock.release()
                self._log.debug("Hook complete: %s", path)
                exec_deferred.callback(True)

            if self._observer:
                self._observer(path)

    @inlineCallbacks
    def stop(self):
        """Stop hook executions.

        Returns a deferred that fires when the executor has stopped.
        """
        assert self._running, "Already stopped"
        yield self._run_lock.acquire()
        self._running = False
        self._executions.put(self.STOP)
        self._run_lock.release()

    def set_observer(self, observer):
        """Set a callback hook execution observer.

        The callback receives a single parameter, the path to the hook,
        and is invoked after the hook has been executed.
        """
        self._observer = observer

    def get_hook_context(self, client_id):
        """Retrieve the context of the currently executing hook.

        This serves as the integration point with the hook api server,
        which utilizes this function to retrieve a hook context for
        a given client. Since we're serializing execution its effectively
        a constant lookup to the currently executing hook's context.
        """
        return self._hook_context

    def get_hook_path(self, hook_path):
        """Retrieve a hook path. We use this to enable debugging.

        :param hook_path:  The requested hook path to execute.

        If the executor is in debug mode, a path to a debug hook
        is returned.
        """
        hook_name = os.path.basename(hook_path)

        # Cleanup/Release any previous hook debug scripts.
        if self._debug_hook_file:
            self._debug_hook_file.close()
            self._debug_hook_file = None

        # Check if debug is active, if not use the requested hook.
        if not self._debug_hook_names:
            return hook_path

        # Check if its a hook we want to debug
        found = False
        for debug_name in self._debug_hook_names:
            if fnmatch.fnmatch(hook_name, debug_name):
                found = True

        if not found:
            return hook_path

        # Setup a debug hook script.
        debug_hook = DEBUG_HOOK_TEMPLATE % hook_name
        self._debug_hook_file = tempfile.NamedTemporaryFile()
        self._debug_hook_file.write(debug_hook)
        self._debug_hook_file.flush()
        os.chmod(self._debug_hook_file.name, 0700)

        return self._debug_hook_file.name

    def set_debug(self, hook_names):
        """Set some hooks to be debugged. Also used to clear debug.

        :param hook_names:
        """
        if hook_names is not None and not isinstance(hook_names, list):
            raise AssertionError("Invalid hook names %r" % (hook_names))
        self._debug_hook_names = hook_names

    def __call__(self, invoker, hook_path):
        """Schedule a hook for execution.

        Returns a deferred that fires when the hook has been executed.
        """
        exec_deferred = Deferred()
        self._executions.put(
            (invoker, hook_path, exec_deferred))
        return exec_deferred
