"""
Protocol

Twisted AMP protocol used between the UnitAgent (via the
ensemble/hooks/invoker template) and client scripts invoked hooks on
behalf of formula authors.

Interactions with the server happen through an exchange of
commands. Each interaction with the UnitAgent is coordinated through
the use of a single command.

These commands have there concrete implementation relative to server
state in the UnitAgentServer class. The utility methods in
UnitAgentClient provide a synchronous interface for scripts derived
from ensemble.hooks.cli to expose to scripts.

To extend the system with additional command the following pattern is
used.

- Author a new BaseCommand subclass outlining the arguments and returns
the Command neeeds.

- Implement a responder for that command in UnitAgentServer returning
  a dict with the response agreed upon by the new Command

- Implement a client side callable in UnitAgentClient which handles
  any pre-wire data marshaling (with the goal of mapping to the
  Command objects contract) and return a result after waiting for any
  asynchronous actions to complete.

UnitAgentClient and UnitAgentServer act as the client and server sides
of an RPC interface. Due to this they have a number of arguments in
common which are documented here.

arguments:

`client_id` -- Client specifier identifying a client to the server
side thus connecting it with an ensemble.state.hook.HookContent (str)

`unit_name` -- String of the name of the unit being queried or
manipulated.


"""
import json

from twisted.internet import defer
from twisted.internet import protocol
from twisted.protocols import amp

from ensemble.errors import EnsembleError
from ensemble.state.errors import UnitRelationStateNotFound


class NoSuchUnit(EnsembleError):
    """
    The requested Unit Name wasn't found
    """
    # Amp Currently cannot construct the 3 required arguments for
    # UnitRelationStateNotFound. This captures the error message in
    # a way that can pass over the wire
    pass


class BaseCommand(amp.Command):
    errors = {NoSuchUnit: "NoSuchUnit",
              UnitRelationStateNotFound: "UnitRelationStateNotFound"}


# All the commands below this point should be documented in the
# specification specifications/unit-agent-hooks
class GetCommand(BaseCommand):
    commandName = "get"
    arguments = [("client_id", amp.String(),),
                 ("unit_name", amp.String(),),
                 ("setting_name", amp.String())]
    response = [("data", amp.String())]


class SetCommand(BaseCommand):
    commandName = "set"
    arguments = [("client_id", amp.String(),),
                 ("json_blob", amp.String())]
    response = []


class ListRelationsCommand(BaseCommand):
    # comma delimited strings
    arguments = [("client_id", amp.String(),)]
    response = [("members", amp.String())]


class LogCommand(BaseCommand):
    arguments = [("level", amp.Integer()),
                 ("message", amp.String())]
    response = []


class UnitAgentServer(amp.AMP, object):
    """
    Protocol used by the UnitAgent to provide a server side to CLI
    tools
    """

    def connectionMade(self):
        """Inform the factory a connection was made.
        """
        super(UnitAgentServer, self).connectionMade()
        self.factory.connectionMade(self)

    @GetCommand.responder
    @defer.inlineCallbacks
    def get(self, client_id, unit_name, setting_name):
        """Get settings from a state.hook.HookContext

        :param settings_name: optional setting_name (str) indicating that
        the client requested a single value only.

        """
        context = self.factory.get_context(client_id)
        try:
            if setting_name:
                data = yield context.get_value(unit_name, setting_name)
            else:
                data = yield context.get(unit_name)
        except UnitRelationStateNotFound, e:
            raise NoSuchUnit(str(e))
        defer.returnValue(dict(data=json.dumps(data)))

    @SetCommand.responder
    @defer.inlineCallbacks
    def set(self, client_id, json_blob):
        """Set values into state.hook.HookContext.

        :param json_blob: a JSON serialized string of a dict that will
        contain the delta of settings to be applied to a unit_name.
        """
        data = json.loads(json_blob)
        context = yield self.factory.get_context(client_id)
        for k, v in data.items():
            yield context.set_value(k, v)
        defer.returnValue({})

    @ListRelationsCommand.responder
    @defer.inlineCallbacks
    def list_relations(self, client_id):
        """Lists the members of a Relation.
        """
        context = yield self.factory.get_context(client_id)
        members = yield context.get_members()
        defer.returnValue(dict(members=" ".join(members)))

    @LogCommand.responder
    @defer.inlineCallbacks
    def log(self, level, message):
        """Log a message from the hook with the UnitAgent.

        :param level: A python logging module log level integer
        indicating the level the message should be logged at.

        :param message: A string containing the message to be logged.
        """
        yield self.factory.log(level, message)
        defer.returnValue({})


class UnitAgentClient(amp.AMP, object):
    """
    Helper used by the CLI tools to call the UnitAgentServer protocol run in
    the UnitAgent.
    """
    @defer.inlineCallbacks
    def get(self, client_id, unit_name, setting_name):
        """ See UnitAgentServer.get
        """
        if not setting_name:
            setting_name = ""
        result = yield self.callRemote(GetCommand,
                                       client_id=client_id,
                                       unit_name=unit_name,
                                       setting_name=setting_name)
        defer.returnValue(json.loads(result["data"]))

    @defer.inlineCallbacks
    def set(self, client_id, data):
        """Set relation settings for unit_name

        :param data: Python dict applied as a delta hook settings

        """
        json_blob = json.dumps(data)
        yield self.callRemote(SetCommand,
                              client_id=client_id,
                              json_blob=json_blob)
        defer.returnValue(None)

    @defer.inlineCallbacks
    def list_relations(self, client_id):
        result = yield self.callRemote(ListRelationsCommand,
                                       client_id=client_id)
        members = result["members"].split()
        defer.returnValue(members)

    @defer.inlineCallbacks
    def log(self, level, message):
        if isinstance(message, (list, tuple)):
            message = " ".join(message)

        result = yield self.callRemote(LogCommand,
                                       level=level,
                                       message=message)
        defer.returnValue(result)


class UnitSettingsFactory(protocol.ServerFactory, object):
    protocol = UnitAgentServer

    def __init__(self, context_provider, log=None):
        """ Factory to be used by the server for communications.

        :param context_provider: Callable(client_id) returning an
        ensemble.state.hook.HookContext. A given client_id will map to
        a single HookContext.

        :param log: When not None a python.logging.Logger object. The log
        is usually managed by the UnitAgent and is passed through the
        factory.

        """
        self.context_provider = context_provider
        self._log = log
        self.onMade = defer.Deferred()

    def get_context(self, client_id):
        return self.context_provider(client_id)

    def log(self, level, message):
        if self._log is not None:
            self._log(level, message)

    def connectionMade(self, protocol):
        if self.onMade:
            self.onMade.callback(protocol)
            self.onMade = None
