from StringIO import StringIO
import logging

from twisted.internet import protocol, defer, error

from ensemble.hooks.protocol import (UnitAgentClient, UnitAgentServer,
                                     UnitSettingsFactory, NoSuchUnit)
from ensemble.lib.testing import TestCase
from ensemble.lib.twistutils import gather_results
from ensemble.state.errors import UnitRelationStateNotFound


def _loseAndPass(err, proto):
    # be specific, pass on the error to the client.
    err.trap(error.ConnectionLost, error.ConnectionDone)
    del proto.connectionLost
    proto.connectionLost(err)


class UnitAgentServerMock(UnitAgentServer):
    def _get_data(self):
        return self.factory.data

    def _set_data(self, dictlike):
        """
        protected method used in testing to rewrite internal data state
        """
        self.factory.data = dictlike

    data = property(_get_data, _set_data)

    def _set_members(self, members):
        # replace the content of the current list
        self.factory.members = members


class UnitSettingsFactoryLocal(UnitSettingsFactory):
    """
    For testing a UnitSettingsFactory with local storage. Loosely
    mimics a HookContext
    """
    protocol = UnitAgentServerMock

    def __init__(self):
        super(UnitSettingsFactoryLocal, self).__init__(self.context_provider)
        self.data = {}
        self.members = []
        self._agent_io = StringIO()

    def log(self, level, message):
        print >> self._agent_io, "%s %s" % (level, message)

    def context_provider(self, client_id):
        return self

    def get_value(self, unit_name, setting_name):
        return self.data[unit_name][setting_name]

    def get(self, unit_name):
        # Currently this is cheating as the real impl can
        if unit_name not in self.data:
            # raise it with fake data
            raise UnitRelationStateNotFound("mysql/1",
                                            "server",
                                            unit_name)

        return self.data[unit_name]

    def get_members(self):
        return self.members

    def set_value(self, key, value):
        self.data.setdefault(self._unit_name, {})[key] = value

    def set(self, blob):
        self.data[self._unit_name] = blob

    def _set_unit_name(self, unit_name):
        self._unit_name = unit_name


class LiveFireBase(TestCase):
    """
    Utility for connected reactor-using tests.
    """
    def _listen_server(self, addr):
        from twisted.internet import reactor

        self.server_factory = UnitSettingsFactoryLocal()
        self.server_socket = reactor.listenUNIX(addr, self.server_factory)
        self.addCleanup(self.server_socket.stopListening)
        return self.server_socket

    def _connect_client(self, addr):
        from twisted.internet import reactor
        d = protocol.ClientCreator(
            reactor, self.client_protocol).connectUNIX(addr)
        return d

    def setUp(self):
        """
        Create an amp server and connect a client to it.
        """
        super(LiveFireBase, self).setUp()
        sock = self.makeFile()

        self._listen_server(sock)
        on_client_connect = self._connect_client(sock)

        def getProtocols(results):
            [(_, client), (_, server)] = results
            self.client = client
            self.server = server

        dl = defer.DeferredList([on_client_connect,
                                 self.server_factory.onMade])
        return dl.addCallback(getProtocols)

    def tearDown(self):
        """
        Cleanup client and server connections, and check the error got at
        C{connectionLost}.
        """
        L = []

        for conn in self.client, self.server:
            if conn.transport is not None:
                # depend on amp's function connection-dropping behavior
                d = defer.Deferred().addErrback(_loseAndPass, conn)
                conn.connectionLost = d.errback
                conn.transport.loseConnection()
                L.append(d)

        super(LiveFireBase, self).tearDown()
        return gather_results(L)


class TestCommunications(LiveFireBase):
    """
   Verify that client and server can communicate with the proper
    protocol.
    """
    client_protocol = UnitAgentClient
    server_protocol = UnitAgentServer
    # short timeout, these are not long running requests when
    # correct
    timeout = 1.0

    @defer.inlineCallbacks
    def test_get_command(self):
        # provide fake data to the server so the client can test it for
        # verification
        self.server.data = dict(test_node=dict(a="b", foo="bar"))
        self.assertIn("test_node", self.server.factory.data)

        data = yield self.client.get("client_id", "test_node", "a")
        self.assertEquals(data, "b")

        data = yield self.client.get("client_id", "test_node", "foo")
        self.assertEquals(data, "bar")

        # A request for <empty_string> asks for all the settings
        data = yield self.client.get("client_id", "test_node", "")
        self.assertEquals(data["a"], "b")
        self.assertEquals(data["foo"], "bar")

    @defer.inlineCallbacks
    def test_get_no_such_unit(self):
        try:
            yield self.client.get("client_id", "missing_unit/99", "")
        except NoSuchUnit:
            pass
        else:
            self.fail("Failed to respond with proper "
                      "UnitRelationStateNotFound")

    @defer.inlineCallbacks
    def test_set_command(self):
        self.assertEquals(self.server.data, {})

        # for testing mock the context being stored in the factory
        self.server_factory._set_unit_name("test_node")

        result = yield self.client.set("client_id",
                                       dict(a="b", foo="bar"))

        # set returns nothing
        self.assertEqual(result, None)

        # verify the data exists in the server now
        self.assertTrue(self.server.data)
        self.assertEquals(self.server.data["test_node"]["a"], "b")
        self.assertEquals(self.server.data["test_node"]["foo"], "bar")

    @defer.inlineCallbacks
    def test_list_relations(self):
        self.server.factory.relation_type = "peer"
        self.server._set_members(["riak/1", "riak/2"])

        members = yield self.client.list_relations("client_id")
        self.assertIn("riak/1", members)
        self.assertIn("riak/2", members)

    @defer.inlineCallbacks
    def test_log_command(self):
        self.save_logging()
        # This is the default calling convention from clients
        yield self.client.log(logging.WARNING, ["This", "is", "a", "WARNING"])
        yield self.client.log(logging.INFO, "This is INFO")
        yield self.client.log(logging.CRITICAL, ["This is CRITICAL"])

        self.assertInDefaultLog("%d This is a WARNING" % logging.WARNING,
                                self.server_factory._agent_io)
        self.assertInDefaultLog("%d This is INFO" % logging.INFO,
                                self.server_factory._agent_io)
        self.assertInDefaultLog("%d This is CRITICAL" % logging.CRITICAL,
                                self.server_factory._agent_io)
