from StringIO import StringIO
import json
import logging
import os
import sys
import yaml

from twisted.internet import defer

import ensemble

from ensemble import errors
from ensemble.hooks import invoker
from ensemble.hooks.protocol import UnitSettingsFactory
from ensemble.lib.twistutils import get_module_directory
from ensemble.state import hook
from ensemble.state.endpoint import RelationEndpoint
from ensemble.state.tests.test_relation import RelationTestBase


class MockUnitAgent(object):
    """
    Pretends to implement the client state cache, and the UA hook
    socket
    """
    def __init__(self, client, socket_path):
        self.client = client
        self.socket_path = socket_path
        self._clients = {}  # client_id -> HookContext

        self._agent_log = logging.getLogger("unit-agent")
        self._agent_io = StringIO()
        handler = logging.StreamHandler(self._agent_io)
        self._agent_log.addHandler(handler)

        self.server_listen()

    def get_invoker(self, relation_name, change_type, unit_name,
                    unit_relation,
                    client_id="client_id"):
        change = hook.RelationChange(relation_name, change_type, unit_name)
        context = hook.HookContext(self.client, unit_relation, change)
        self._clients[client_id] = context

        logger = logging.getLogger("hook")
        log_file = StringIO()
        handler = logging.StreamHandler(log_file)
        logger.addHandler(handler)

        exe = invoker.Invoker(context, change,
                              self.get_client_id(),
                              self.socket_path,
                              logger)
        return exe

    def get_client_id(self):
        # simulate associating a client_id with a client connection
        # for later context look up. In reality this would be a mapping.
        return "client_id"

    def get_context(self, client_id):
        return self._clients[client_id]

    def stop(self):
        """
        Stop the process invocation triggering any registered
        cleanup functions
        """
        self.server_socket.stopListening()

    def server_listen(self):
        from twisted.internet import reactor

        # hook context and a logger to the settings factory
        logger = logging.getLogger("unit-agent")
        self.log_file = StringIO()
        handler = logging.StreamHandler(self.log_file)
        logger.addHandler(handler)
        self.server_factory = UnitSettingsFactory(self.get_context, logger.log)

        self.server_socket = reactor.listenUNIX(
            self.socket_path, self.server_factory)


def get_cli_environ_path(*search_path):
    """
    Construct a path environment variable value, that contains
    the ensemble bin directory.

    @param search_path: additional directories to put on PATH
    """
    search_path = list(search_path)

    # Look for the top level ensemble bin directory and make sure
    # that is available for the client utilities.
    bin_path = os.path.normpath(
        os.path.join(get_module_directory(ensemble), "..", "bin"))

    search_path.append(bin_path)
    search_path.extend(os.environ.get("PATH", "").split(":"))

    return ":".join(search_path)


class TestInvoker(RelationTestBase):

    @defer.inlineCallbacks
    def setUp(self):
        yield super(TestInvoker, self).setUp()

        yield self._default_relations()
        self._change_environment()
        self.socket_path = self.makeFile()
        self.ua = MockUnitAgent(self.client, self.socket_path)

    @defer.inlineCallbacks
    def tearDown(self):
        self.ua.stop()
        yield super(TestInvoker, self).tearDown()

    def _change_environment(self):
        test_hook_path = os.path.join(
            os.path.abspath(
                os.path.dirname(__file__)).replace("/_trial_temp", ""),
            "hooks")
        self.change_environment(
            PATH=get_cli_environ_path(test_hook_path),
            ENSEMBLE_UNIT_NAME="mysql/1")

    @defer.inlineCallbacks
    def _default_relations(self):
        wordpress_ep = RelationEndpoint(
            "wordpress", "client-server", "", "client")
        mysql_ep = RelationEndpoint(
            "mysql", "client-server", "", "server")
        self.wordpress_states = \
            yield self.add_relation_service_unit_from_endpoints(
                wordpress_ep, mysql_ep)
        self.mysql_states = yield self.add_opposite_service_unit(
            self.wordpress_states)
        self.relation = self.mysql_states["unit_relation"]

    def get_test_hook(self, hook):
        dirname = os.path.dirname(__file__)
        abspath = os.path.abspath(dirname)
        hook_file = os.path.join(abspath, "hooks", hook)

        if not os.path.exists(hook_file):
            # attempt to find it via sys_path
            for p in sys.path:
                hook_file = os.path.normpath(
                    os.path.join(p, dirname, "hooks", hook))
                if os.path.exists(hook_file):
                    return hook_file
            raise IOError("%s doesn't exist" % hook_file)
        return hook_file

    def test_environment(self):
        """
        there are various way to manipulate the calling environment
        from Invoker objects.
        """
        exe = self.ua.get_invoker("database", "add", "mysql/1", self.relation)
        exe.environment.update(dict(FOO="bar"))
        env = exe.get_environment()

        # these come from the init argument
        self.assertEqual(env["ENSEMBLE_AGENT_SOCKET"], self.socket_path)
        self.assertEqual(env["ENSEMBLE_CLIENT_ID"], "client_id")

        # this comes from updating the Invoker.environment
        self.assertEqual(env["FOO"], "bar")

        # and this comes from the unit agent passing through its environment
        self.assertTrue(env["PATH"])
        self.assertEqual(env["ENSEMBLE_UNIT_NAME"], "mysql/1")

    def test_missing_hook(self):
        self.save_logging()

        exe = self.ua.get_invoker("database", "add", "mysql/1", self.relation)
        self.failUnlessRaises(errors.FileNotFound, exe, "hook-missing")

    def test_noexec_hook(self):
        self.save_logging()

        exe = self.ua.get_invoker("database", "add", "mysql/1",
                                  self.relation)
        hook = self.get_test_hook("noexec-hook")
        error = self.failUnlessRaises(errors.FormulaError, exe, hook)

        self.assertEqual(error.path, hook)
        self.assertIn(error.message, "hook is not executable")

    @defer.inlineCallbacks
    def test_spawn_success(self):
        """
        validate hook with success from exit
        """
        self.save_logging()
        exe = self.ua.get_invoker("database", "add", "mysql/1", self.relation)
        result = yield exe(self.get_test_hook("success-hook"))
        self.assertEqual(result, 0)
        self.assertInDefaultLog("WIN")
        self.assertInDefaultLog("exit code 0")

    @defer.inlineCallbacks
    def test_spawn_fail(self):
        """
        validate hook with fail from exit
        """
        self.save_logging()
        self.reset_logging()
        exe = self.ua.get_invoker("database", "add", "mysql/1", self.relation)
        d = exe(self.get_test_hook("fail-hook"))
        result = yield self.assertFailure(d, errors.FormulaInvocationError)
        self.assertEqual(result.exit_code, 1)
        # ERROR indicate the level name, we are checking that the
        # proper level was logged here
        self.assertInDefaultLog("ERROR")
        # and the message
        self.assertInDefaultLog("FAIL")
        self.assertInDefaultLog("exit code 1")

    def test_path_setup(self):
        """
        validate that the path allows finding the executable
        """
        from twisted.python.procutils import which
        exe = which("relation-get")
        self.assertTrue(exe)
        self.assertTrue(exe[0].endswith("relation-get"))

    @defer.inlineCallbacks
    def test_spawn_cli_get_hook(self):
        """
        Validate the get hook can transmit values to the hook
        """
        hook_log = self.capture_logging("hook")
        exe = self.ua.get_invoker("database", "add", "mysql/1",
                                  self.relation,
                                  client_id="client_id")

        # Populate and verify some data we will
        # later extract with the hook
        context = self.ua.get_context("client_id")
        expected = {"a": "b", "c": "d"}
        yield context.set(expected)
        data = yield context.get("mysql/0")

        self.assertEqual(expected, data)

        # invoke the hook and process the results
        # verifying they are expected
        result = yield exe(self.get_test_hook("get-hook"))
        self.assertEqual(result, 0)
        data = hook_log.getvalue()
        self.assertEqual(json.loads(data), expected)

    @defer.inlineCallbacks
    def test_spawn_cli_get_value_hook(self):
        """
        Validate the get hook can transmit values to the hook
        """
        hook_log = self.capture_logging("hook")
        exe = self.ua.get_invoker("database", "add", "mysql/1",
                                  self.relation,
                                  client_id="client_id")

        # Populate and verify some data we will
        # later extract with the hook
        context = self.ua.get_context("client_id")
        expected = {"name": "rabbit"}
        yield context.set(expected)
        data = yield context.get("mysql/0")

        self.assertEqual(expected, data)

        # invoke the hook and process the results
        # verifying they are expected
        result = yield exe(self.get_test_hook("get-value-hook"))
        self.assertEqual(result, 0)
        data = hook_log.getvalue()
        self.assertEqual("rabbit", json.loads(data))

    @defer.inlineCallbacks
    def test_spawn_cli_set_hook(self):
        """
        Validate the set hook can set values in zookeeper.
        """
        exe = self.ua.get_invoker("database", "add", "mysql/1",
                                  self.relation,
                                  client_id="client_id")

        # Invoke the hook and process the results verifying they are expected
        result = yield exe(self.get_test_hook("set-hook"))
        self.assertEqual(result, 0)

        # Verify the context was flushed to zk
        zk_data = yield self.relation.get_data()
        self.assertEqual({"a": "b", "c": "d"}, yaml.load(zk_data))

    @defer.inlineCallbacks
    def test_logging(self):
        self.save_logging()
        self.reset_logging()
        exe = self.ua.get_invoker("database", "add", "mysql/1", self.relation)

        # The echo hook will echo out the value
        # it will also echo to stderr the ERROR variable
        message = "This is only a test"
        error = "All is full of fail"

        exe.environment["MESSAGE"] = message
        exe.environment["ERROR"] = error
        # of the MESSAGE variable
        result = yield exe(self.get_test_hook("echo-hook"))
        self.assertEqual(result, 0)

        self.assertInDefaultLog(message)
        # The 'error' was sent via ensemble-log
        # to the UA. Our test UA has a fake log stream
        # which we can check now
        output = self.ua.log_file.getvalue()
        self.assertInDefaultLog(error)

        assert message not in output, """Log includes unintended messages"""

    @defer.inlineCallbacks
    def test_spawn_cli_list_hook_smart(self):
        """
        Validate the get hook can transmit values to the hook
        """
        self.save_logging()
        self.reset_logging()
        exe = self.ua.get_invoker("database", "add", "mysql/1",
                                  self.relation,
                                  client_id="client_id")

        # Populate and verify some data we will
        # later extract with the hook
        context = self.ua.get_context("client_id")

        # directly manipulate the context to the expected list of
        # members
        expected = ["alpha/0", "beta/0"]
        context._members = expected

        # invoke the hook and process the results
        # verifying they are expected
        exe.environment["FORMAT"] = "smart"
        result = yield exe(self.get_test_hook("list-hook"))

        self.assertEqual(result, 0)
        self.assertInDefaultLog("alpha/0\nbeta/0\n")

    @defer.inlineCallbacks
    def test_spawn_cli_list_hook_eval(self):
        """
        Validate the get hook can transmit values to the hook
        """
        self.save_logging()
        self.reset_logging()
        exe = self.ua.get_invoker("database", "add", "mysql/1",
                                  self.relation,
                                  client_id="client_id")

        # Populate and verify some data we will
        # later extract with the hook
        context = self.ua.get_context("client_id")

        # directly manipulate the context to the expected list of
        # members
        expected = ["alpha/0", "beta/0"]
        context._members = expected

        # invoke the hook and process the results
        # verifying they are expected
        exe.environment["FORMAT"] = "eval"
        result = yield exe(self.get_test_hook("list-hook"))

        self.assertEqual(result, 0)
        self.assertInDefaultLog("alpha/0 beta/0")

    @defer.inlineCallbacks
    def test_spawn_cli_list_hook_json(self):
        """
        Validate the get hook can transmit values to the hook
        """
        self.save_logging()
        self.reset_logging()
        exe = self.ua.get_invoker("database", "add", "mysql/1",
                                  self.relation,
                                  client_id="client_id")

        # Populate and verify some data we will
        # later extract with the hook
        context = self.ua.get_context("client_id")

        # directly manipulate the context to the expected list of
        # members
        expected = ["alpha/0", "beta/0"]
        context._members = expected

        # invoke the hook and process the results
        # verifying they are expected
        exe.environment["FORMAT"] = "json"
        result = yield exe(self.get_test_hook("list-hook"))

        self.assertEqual(result, 0)
        self.assertInDefaultLog('["alpha/0", "beta/0"]')
