import yaml
import zookeeper

from twisted.internet.defer import inlineCallbacks, returnValue

from txzookeeper.utils import retry_change

from ensemble.state.base import StateBase
from ensemble.state.errors import UnitRelationStateNotFound, StateError


class RelationChange(object):
    """Encapsulation of relation change variables passed to hook.
    """

    def __init__(self, relation_name, change_type, unit_name):
        self._relation_name = relation_name
        self._change_type = change_type
        self._unit_name = unit_name

    @property
    def relation_name(self):
        return self._relation_name

    @property
    def change_type(self):
        return self._change_type

    @property
    def unit_name(self):
        return self._unit_name


class HookContext(StateBase):
    """A hook execution data cache and write buffer of relation settings.

    Performs caching of any relation settings examined by the
    hook. Also buffers all writes till the flush method is invoked.
    """

    def __init__(self, client, unit_relation, change, members=None):
        """
        @param unit_relation: The unit relation state associated to the hook.
        @param change: A C{RelationChange} instance.
        """
        # Zookeeper client.
        super(HookContext, self).__init__(client)
        self._unit_relation = unit_relation

        # The current change we're being executed against.
        self._change = change

        # A cache of retrieved nodes.
        self._node_cache = {}

        # A cache of node names to node ids.
        self._name_cache = {}

        # A cache of writes
        self._write_buffer = None

        # We keep a pristine copy of the original settings of the
        # associated context unit settings, to be able to detect
        # deletes done on the source.
        self._pristine_settings = None

        # A cache of related units in the relation.
        self._members = members

        # Topology for resolving name<->id of units.
        self._topology = None

    @inlineCallbacks
    def _resolve_id(self, unit_id):
        """Resolve a unit id to a unit name."""
        if self._topology is None:
            self._topology = yield self._read_topology()
        unit_name = self._topology.get_service_unit_name_from_id(unit_id)
        returnValue(unit_name)

    @inlineCallbacks
    def _resolve_name(self, unit_name):
        """Resolve a unit name to a unit id with caching."""
        if unit_name in self._name_cache:
            returnValue(self._name_cache[unit_name])
        if self._topology is None:
            self._topology = yield self._read_topology()
        unit_id = self._topology.get_service_unit_id_from_name(unit_name)
        self._name_cache[unit_name] = unit_id
        returnValue(unit_id)

    def _settings_path(self, unit_id):
        return "/relations/%s/settings/%s" % (
            self._unit_relation.internal_relation_id, unit_id)

    @inlineCallbacks
    def get_members(self):
        """Gets the related unit members of the relation with caching."""
        if self._members is not None:
            returnValue(self._members)

        container = yield self._unit_relation.get_related_unit_container()
        unit_ids = yield self._client.get_children(container)
        if self._unit_relation.internal_unit_id in unit_ids:
            unit_ids.remove(self._unit_relation.internal_unit_id)

        members = []
        for unit_id in unit_ids:
            unit_name = yield self._resolve_id(unit_id)
            members.append(unit_name)

        self._members = members
        returnValue(members)

    @inlineCallbacks
    def get(self, unit_name):
        """Get the relation settings for a unit.

        Returns the settings as a dictionary.
        """
        unit_id = yield self._resolve_name(unit_name)

        # If we're fetching our own values, be sure to account
        # for any local modifications.
        if (unit_id == self._unit_relation.internal_unit_id
            and self._write_buffer is not None):
            returnValue(dict(self._write_buffer))

        # Check if we have a cached value for the unit.
        if unit_name in self._node_cache:
            returnValue(dict(self._node_cache[unit_name]))

        # Fetch, cache, and return the settings data.
        path = self._settings_path(unit_id)
        try:
            data, stat = yield self._client.get(path)
        except zookeeper.NoNodeException:
            raise UnitRelationStateNotFound(
                self._unit_relation.internal_relation_id,
                self._change.relation_name,
                unit_name)

        if data:
            data = yaml.load(data)
            if not isinstance(data, dict):
                msg = "Invalid relation data for unit %r, should be dict" % (
                    unit_name)
                raise StateError(msg)
        else:
            data = {}

        self._node_cache[unit_name] = data
        returnValue(dict(data))

    @inlineCallbacks
    def get_value(self, unit_name, key):
        """Get a relation setting value for a unit."""
        settings = yield self.get(unit_name)
        if not settings:
            returnValue("")
        returnValue(settings.get(key, ""))

    @inlineCallbacks
    def set(self, data):
        """Set the relation settings for a unit.

        @param data: A dictionary containing the settings.

        **Warning**, this method will replace existing values for the
        unit relation with those from the ``data`` dictionary.
        """
        if not isinstance(data, dict):
            raise TypeError("A dictionary is required.")

        # Fetch the pristine copy for deletion detection during flush.
        if self._pristine_settings is None:
            unit_name = yield self._resolve_id(
                self._unit_relation.internal_unit_id)
            pristine = yield self.get(unit_name)
            self._pristine_settings = pristine

        self._write_buffer = data

    @inlineCallbacks
    def set_value(self, key, value):
        """Set a relation value for a unit."""
        if self._write_buffer is None:
            # Utilize the context get instead of the state get to take
            # advantage of the cache, and dict deserialization.
            unit_name = yield self._resolve_id(
                self._unit_relation.internal_unit_id)
            self._write_buffer = yield self.get(unit_name)
            self._pristine_settings = dict(self._write_buffer)
        self._write_buffer[key] = value

    def has_read(self, unit_name):
        """Has the context been used to access the settings of the unit.
        """
        return unit_name in self._node_cache

    @inlineCallbacks
    def flush(self):
        """Flush all writes to the unit settings.

        A flush will attempt to intelligently merge values modified on
        the context to the current state of the underlying settings
        node.  It supports externally modified or deleted values that
        are unchanged on the context, to be preserved.
        """
        # If there are no changes, we're done.
        if not self._write_buffer:
            returnValue(None)

        # Make a copy to prevent concurrent modification.
        write_buffer = self._write_buffer
        pristine_settings = self._pristine_settings or {}
        self._write_buffer = None
        self._pristine_settings = None

        def apply_changes(content, stat):
            """Merge the context settings with the node state."""
            data = yaml.load(content)
            if not data:
                return yaml.safe_dump(write_buffer)

            # Detect deletions in the current, so we don't resurrect
            # unmodified dead values.
            if pristine_settings:
                previous_keys = set(pristine_settings)
                current_keys = set(data.keys())
                removed = previous_keys - current_keys
                for rk in removed:
                    old_value = pristine_settings[rk]
                    new_value = write_buffer[rk]
                    if new_value is None:
                        continue
                    if old_value == new_value:
                        del write_buffer[rk]

            for key, new_value in write_buffer.items():
                if new_value is None:
                    if key in data:
                        del data[key]
                elif key not in pristine_settings:
                    # If its a new value, set it.
                    data[key] = new_value
                else:
                    # For existing values, set if its been modified.
                    if new_value != pristine_settings[key]:
                        data[key] = new_value

            return yaml.safe_dump(data)

        # Apply the change till it takes.
        yield retry_change(
            self._client,
            self._settings_path(
                self._unit_relation.internal_unit_id),
            apply_changes)
