import errno
import socket
import time
import zookeeper

from Queue import Queue

from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from twisted.internet.threads import deferToThread
from txzookeeper import ZookeeperClient

from ensemble.lib.testing import TestCase
from ensemble.state.utils import PortWatcher, remove_tree, get_open_port
from ensemble.tests.common import get_test_zookeeper_address


class PortWait(TestCase):

    def bind_port(self, delay, port=0):
        """Create a fake server for testing.

        Binds socket to `port` for `delay` seconds, then listens. If
        `port` is 0, then it is assigned by the system.  Returns a
        pair of deferreds: completed, port

        The `completed` deferred is called back when the socket is
        closed. The port deferred has the port number, which is useful
        if a system-assigned port is requested.
        """
        # used to communicate the port number
        port_queue = Queue()

        def bind_port_sync():
            sock = socket.socket()
            sock.bind(("127.0.0.1", port))
            sock.listen(1)
            port_queue.put(sock.getsockname()[1])
            time.sleep(delay)
            sock.close()

        return deferToThread(bind_port_sync), deferToThread(port_queue.get)

    @inlineCallbacks
    def test_existing_connect(self):
        """Test watch fires asap if a real server is then available."""
        server_deferred, port_deferred = self.bind_port(0.2)
        port = yield port_deferred
        yield self.sleep(0.1)
        yield PortWatcher("127.0.0.1", port, 1).async_wait()
        yield server_deferred

    @inlineCallbacks
    def test_connect(self):
        """Test watch fires soon after a real server becomes available.

        0.5 second for the approximation is based on the polling
        interval of PortWatcher."""
        now = time.time()
        reactor.callLater(0.7, self.bind_port, 1, 22181)
        yield PortWatcher("127.0.0.1", 22181, 1.5).async_wait()
        self.failUnlessApproximates(time.time() - now, 0.7, 0.5)

    def test_wait_until_timeout_raises_timeout(self):
        """If the timeout is exceeded, a socket.timeout error is raised."""
        self.assertRaises(
            socket.timeout, PortWatcher("localhost", 800, 0).sync_wait)

    def test_wait_stops_when_watcher_stopped(self):
        """
        If the watcher is stopped, no more attempts are made to attempt
        connect to the socket.
        """
        watcher = PortWatcher("127.0.0.1", 800, 30)

        mock_socket = self.mocker.patch(socket.socket)
        sleep = self.mocker.replace("time.sleep")

        mock_socket.connect(("127.0.0.1", 800))
        self.mocker.throw(socket.error(errno.ECONNREFUSED))
        sleep(0.5)
        self.mocker.call(watcher.stop)
        self.mocker.replay()
        self.assertEquals(watcher.sync_wait(), None)

    def test_unknown_socket_error_raises(self):
        """Unknown socket errors are thrown to the caller."""
        mock_socket = self.mocker.patch(socket.socket)
        mock_socket.connect(("127.0.0.1", 465))
        self.mocker.throw(socket.error(2000))
        self.mocker.replay()
        watcher = PortWatcher("127.0.0.1", 465, 5)
        self.assertRaises(socket.error, watcher.sync_wait)

    def test_unknown_error_raises(self):
        """Unknown errors are thrown to the caller."""
        mock_socket = self.mocker.patch(socket.socket)
        mock_socket.connect(("127.0.0.1", 465))
        self.mocker.throw(SyntaxError())
        self.mocker.replay()
        watcher = PortWatcher("127.0.0.1", 465, 5)
        self.assertRaises(SyntaxError, watcher.sync_wait)

    def test_on_connect_returns(self):
        """On a successful connect, the function returns None."""
        mock_socket = self.mocker.patch(socket.socket)
        mock_socket.connect(("127.0.0.1", 800))
        self.mocker.result(True)
        mock_socket.close()

        self.mocker.replay()
        watcher = PortWatcher("127.0.0.1", 800, 30)
        self.assertEqual(watcher.sync_wait(), True)

    @inlineCallbacks
    def test_listen(self):
        """Test with a real socket server watching for port availability."""
        bind_port_time = 0.7
        now = time.time()
        server_deferred, port_deferred = self.bind_port(bind_port_time)
        port = yield port_deferred
        yield PortWatcher("127.0.0.1", port, 10, True).async_wait()
        self.failUnlessApproximates(time.time() - now, bind_port_time, 0.5)
        yield server_deferred

    @inlineCallbacks
    def test_listen_nothing_there(self):
        """Test with a real socket server watching for port availability.

        This should result in the watch returning almost immediately,
        since nothing is (or should be - dangers of real testing)
        holding this port."""
        now = time.time()
        yield PortWatcher("127.0.0.1", 22181, 10, True).async_wait()
        self.failUnlessApproximates(time.time() - now, 0, 0.1)

    def test_listen_wait_stops_when_watcher_stopped(self):
        """
        If the watcher is stopped, no more attempts are made to attempt
        to listen to the socket.
        """
        watcher = PortWatcher("127.0.0.1", 800, 30, True)

        mock_socket = self.mocker.patch(socket.socket)
        sleep = self.mocker.replace("time.sleep")

        mock_socket.bind(("127.0.0.1", 800))
        self.mocker.throw(socket.error(errno.ECONNREFUSED))
        sleep(0.5)
        self.mocker.call(watcher.stop)
        self.mocker.replay()
        self.assertEquals(watcher.sync_wait(), None)

    def test_listen_unknown_socket_error_raises(self):
        """Unknown socket errors are thrown to the caller."""
        mock_socket = self.mocker.patch(socket.socket)
        mock_socket.bind(("127.0.0.1", 465))
        self.mocker.throw(socket.error(2000))
        self.mocker.replay()
        watcher = PortWatcher("127.0.0.1", 465, 5, True)
        self.assertRaises(socket.error, watcher.sync_wait)

    def test_listen_unknown_error_raises(self):
        """Unknown errors are thrown to the caller."""
        mock_socket = self.mocker.patch(socket.socket)
        mock_socket.bind(("127.0.0.1", 465))
        self.mocker.throw(SyntaxError())
        self.mocker.replay()
        watcher = PortWatcher("127.0.0.1", 465, 5, True)
        self.assertRaises(SyntaxError, watcher.sync_wait)

    def test_listen_on_connect_returns(self):
        """On a successful connect, the function returns None."""
        mock_socket = self.mocker.patch(socket.socket)
        mock_socket.bind(("127.0.0.1", 800))
        self.mocker.result(True)
        mock_socket.close()

        self.mocker.replay()
        watcher = PortWatcher("127.0.0.1", 800, 30, True)
        self.assertEqual(watcher.sync_wait(), True)


class RemoveTreeTest(TestCase):

    @inlineCallbacks
    def setUp(self):
        zookeeper.set_debug_level(0)
        self.client = ZookeeperClient(get_test_zookeeper_address())
        yield self.client.connect()

    @inlineCallbacks
    def test_remove_tree(self):
        yield self.client.create("/zoo")
        yield self.client.create("/zoo/mammals")
        yield self.client.create("/zoo/mammals/elephant")
        yield self.client.create("/zoo/reptiles")
        yield self.client.create("/zoo/reptiles/snake")

        yield remove_tree(self.client, "/zoo")

        children = yield self.client.get_children("/")
        self.assertNotIn("zoo", children)


class OpenPortTest(TestCase):

    def test_get_open_port(self):
        port = get_open_port()
        self.assertTrue(isinstance(port, int))

        sock = socket.socket(
            socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)

        # would raise an error if we got it wrong.
        sock.bind(("127.0.0.1", port))
        sock.listen(1)
        sock.close()
        del sock
