import socket
import errno
import time

from twisted.internet.defer import inlineCallbacks
from twisted.internet.threads import deferToThread


class PortWatcher(object):

    def __init__(self, host, port, timeout, listen=False):
        """Watches a `port` on `host` until available.

        Used with `sync_wait` or `async_wait` methods.

        Times out after `timeout` seconds. Normally the watcher is
        used to determine when a port starts listening for client use,
        but the parameter `listen` may be used to wait for when the
        port can be used by a server (because of previous usage
        without properly closing).
        """
        self._host = host
        self._port = port
        self._timeout = timeout
        self._stop = False
        self._listen = listen

    def stop(self, result=None):
        """Interrupt port watching in its loop."""
        self._stop = True
        return result

    def sync_wait(self):
        """Waits until the port is available, or `socket.error`."""
        until = time.time() + self._timeout
        while time.time() < until and not self._stop:
            sock = socket.socket()
            sock.settimeout(1)
            try:
                if self._listen:
                    sock.bind((self._host, self._port))
                else:
                    sock.connect((self._host, self._port))
            except socket.timeout:
                time.sleep(0.5)
            except socket.error, e:
                if e.args[0] not in (
                    errno.EWOULDBLOCK, errno.ECONNREFUSED, errno.EADDRINUSE):
                    raise
                else:
                    # Sleep, otherwise this code will create useless sockets
                    time.sleep(0.5)
            else:
                sock.close()
                return True

        if self._stop:
            return

        raise socket.timeout("could not connect before timeout")

    def async_wait(self, *ignored):
        """Returns a deferred that is called back on port available.

        An exception is returned through the deferred errback if the
        port wait timed out, or another problem occurs."""
        return deferToThread(self.sync_wait)


@inlineCallbacks
def remove_tree(client, path):
    children = yield client.get_children(path)
    for child in children:
        yield remove_tree(client, "%s/%s" % (path, child))
    yield client.delete(path)


def get_open_port():
    """Get an open port on the machine.
    """
    temp_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    temp_sock.bind(("", 0))
    port = temp_sock.getsockname()[1]
    temp_sock.close()
    del temp_sock
    return port
