import os
import logging

from twisted.internet.defer import (
    inlineCallbacks, DeferredLock)

from ensemble.hooks.invoker import Invoker
from ensemble.hooks.scheduler import HookScheduler
from ensemble.state.hook import RelationChange
from ensemble.state.errors import StopWatcher, UnitRelationStateNotFound

from ensemble.unit.workflow import RelationWorkflowState, RelationWorkflow


HOOK_SOCKET_FILE = ".ensemble.hookcli.sock"

hook_log = logging.getLogger("hook.output")


class UnitLifecycle(object):
    """Manager for a unit lifecycle.

    Primarily used by the workflow interaction, to modify unit behavior
    according to the current unit workflow state and transitions.
    """

    def __init__(self, client, unit, service, unit_path, executor):
        self._client = client
        self._unit = unit
        self._service = service
        self._executor = executor
        self._unit_path = unit_path
        self._relations = {}
        self._running = False
        self._watching = False
        self._run_lock = DeferredLock()
        self._log = logging.getLogger("unit.lifecycle")

    def get_relation_workflow(self, relation_id):
        """Accessor to a unit relation workflow, by relation id.

        Primarily intended for and used by unit tests. Raises
        a KeyError if the relation workflow does not exist.
        """
        return self._relations[relation_id]

    @inlineCallbacks
    def install(self):
        """Invoke the unit's install hook.
        """
        yield self._execute_hook("install")

    @inlineCallbacks
    def start(self):
        """Invoke the start hook, and setup relation watching.
        """
        self._log.debug("pre-start acquire, running:%s", self._running)
        yield self._run_lock.acquire()
        self._log.debug("starting unit lifecycle")
        try:
            # Verify current state
            assert not self._running, "Already started"

            # Execute the start hook
            yield self._execute_hook("start")

            # If we have any existing relations in memory, start them.
            if self._relations:
                self._log.debug("starting relation lifecycles")

            for workflow in self._relations.values():
                # should not transition an
                yield workflow.transition_state("up")

            # Establish a watch on the existing relations.
            if not self._watching:
                self._log.debug("starting service relation watch")
                yield self._service.watch_relation_states(
                    self._on_service_relation_changes)
                self._watching = True

            # Set current status
            self._running = True
        finally:
            self._run_lock.release()
        self._log.debug("started unit lifecycle")

    @inlineCallbacks
    def stop(self):
        """Stop the unit, executes the stop hook, and stops relation watching.
        """
        self._log.debug("pre-stop acquire, running:%s", self._running)
        yield self._run_lock.acquire()
        try:
            # Verify state
            assert self._running, "Already Stopped"

            # Stop relation lifecycles
            if self._relations:
                self._log.debug("starting relation lifecycles")
            for workflow in self._relations.values():
                yield workflow.transition_state("down")

            yield self._execute_hook("stop")

            # Set current status
            self._running = False
        finally:
            self._run_lock.release()
        self._log.debug("stopped unit lifecycle")

    @inlineCallbacks
    def _on_service_relation_changes(self, old_relations, new_relations):
        """Callback for service relation watching.

        The callback is used to manage the unit relation lifecycle in
        accordance with the current relations of the service.

        @param old_relations: Previous service relations for a service. On the
               initial execution, this value is None.
        @param new_relations: Current service relations for a service.
        """
        self._log.debug(
            "services changed old:%s new:%s", old_relations, new_relations)

        # Acquire the run lock, and process the changes.
        yield self._run_lock.acquire()
        try:
            # If the lifecycle is not running, then stop the watcher
            if not self._running:
                self._log.debug("stop service-rel watcher, discarding changes")
                self._watching = False
                raise StopWatcher()

            self._log.debug("processing relations changed")
            yield self._process_service_changes(old_relations, new_relations)
        finally:
            self._run_lock.release()

    @inlineCallbacks
    def _process_service_changes(self, old_relations, new_relations):
        """Add and remove unit lifecycles per the service relations changes.
        """
        # Determine relation delta of global zk state with our memory state.
        new_relations = dict([(service_relation.internal_relation_id,
                               service_relation) for
                              service_relation in new_relations])
        added = set(new_relations.keys()) - set(self._relations.keys())
        removed = set(self._relations.keys()) - set(new_relations.keys())

        # Stop and remove, old ones.

        # Trying to directly transition this causes additional yielding
        # operations, which means that concurrent events for subsequent
        # watch firings will be executed. ie. if the relation
        # is broken, but a subsequent modify comes in for a related unit,
        # it will cause the modify to have a hook execution. To prevent
        # this we stop the lifecycle immediately before executing the
        # transition. see UnitLifecycleTest.test_removed_relation_depart
        for relation_id in removed:
            yield self._relations[relation_id].lifecycle.stop()

        for relation_id in removed:
            workflow = self._relations.pop(relation_id)
            yield workflow.transition_state("departed")

        # Process new relations.
        for relation_id in added:
            service_relation = new_relations[relation_id]
            try:
                unit_relation = yield service_relation.get_unit_state(
                    self._unit)
            except UnitRelationStateNotFound:
                # This unit has not yet been assigned a unit relation state,
                # Go ahead and add one.
                unit_relation = yield service_relation.add_unit_state(
                    self._unit)

            self._log.debug(
                "Starting new relation: %s", service_relation.relation_name)

            workflow = self._get_unit_relation_workflow(unit_relation,
                                                        service_relation)
            # Start it before storing it.
            yield workflow.fire_transition("start")
            self._relations[service_relation.internal_relation_id] = workflow

    def _get_unit_path(self):
        """Retrieve the root path of the unit.
        """
        return self._unit_path

    def _get_unit_relation_workflow(self, unit_relation, service_relation):

        lifecycle = UnitRelationLifecycle(self._client,
                                          unit_relation,
                                          service_relation.relation_name,
                                          self._get_unit_path(),
                                          self._executor)

        state_directory = os.path.abspath(os.path.join(
            self._unit_path, "../../state"))

        workflow = RelationWorkflowState(
            lifecycle, self._client, RelationWorkflow, unit_relation,
            state_directory)

        return workflow

    @inlineCallbacks
    def _execute_hook(self, hook_name):
        """Execute the hook with the given name.
        """
        unit_path = self._get_unit_path()
        hook_path = os.path.join(unit_path, "formula", "hooks", hook_name)
        socket_path = os.path.join(unit_path, HOOK_SOCKET_FILE)

        invoker = Invoker(None, None, "constant", socket_path, hook_log)
        yield self._executor(invoker, hook_path)


class RelationInvoker(Invoker):
    """A relation hook invoker, that populates the environment.
    """

    def get_environment_from_change(self, env, change):
        """Populate environment with relation change information."""
        env["ENSEMBLE_RELATION"] = change.relation_name
        env["ENSEMBLE_REMOTE_UNIT"] = change.unit_name
        env["ENSEMBLE_CHANGE"] = change.change_type
        return env


class UnitRelationLifecycle(object):
    """Unit Relation Lifcycle management.

    Provides for watching related units in a relation, and executing hooks
    in response to changes. The lifecycle is driven by the workflow.

    The Unit relation lifecycle glues together a number of components.
    It controls a watcher that recieves watch events from zookeeper,
    and it controls a hook scheduler which gets fed those events. When
    the scheduler wants to execute a hook, the executor is called with
    the hook path and the hook invoker.

    **Relation hook invocation do not maintain global order or
    determinism across relations**. They only maintain ordering and
    determinism within a relation. A shared scheduler across relations
    would be needed to maintain such behavior.
    """

    def __init__(
        self, client, unit_relation, relation_name, unit_path, executor):

        self._client = client
        self._unit_path = unit_path
        self._relation_name = relation_name
        self._unit_relation = unit_relation
        self._executor = executor
        self._run_lock = DeferredLock()
        self._log = logging.getLogger("unit.relation.lifecycle")
        self._error_handler = None

        self._scheduler = HookScheduler(
            client,
            self._execute_change_hook,
            self._unit_relation,
            self._relation_name)

        self._watcher = None

    @inlineCallbacks
    def _execute_change_hook(self, context, change, hook_name=None):
        """Invoked by the contained HookScheduler, to execute a hook.

        We utilize the HookExecutor to execute the hook, if an
        error occurs, it will be reraised, unless an error handler
        is specified see ``set_hook_error_handler``.
        """
        socket_path = os.path.join(self._unit_path, HOOK_SOCKET_FILE)
        if hook_name is None:
            hook_name = "%s-relation-changed" % self._relation_name

        hook_path = os.path.join(
            self._unit_path, "formula", "hooks", hook_name)

        invoker = RelationInvoker(
            context, change, "constant", socket_path, hook_log)

        yield self._run_lock.acquire()
        self._log.debug("%s: executing change hook", self._relation_name)
        try:
            yield self._executor(invoker, hook_path)
        except Exception, e:
            yield self._run_lock.release()
            self._log.warn("%s: change hook error: %s", self._relation_name, e)

            if not self._error_handler:
                raise
            self._log.info(
                "%s: invoke change hook error handler", self._relation_name)
            # We can't hold the run lock, when we invoke the error handler, or
            # we get a deadlock if the handler manipulates the lifecycle.
            yield self._error_handler(change, e)
        else:
            yield self._run_lock.release()

    def set_hook_error_handler(self, handler):
        """Set an error handler to be invoked if a hook errors.

        The handler should accept one parameter, the exception instance.
        """
        self._error_handler = handler

    @inlineCallbacks
    def start(self):
        """Start watching related units and executing change hooks.
        """
        yield self._run_lock.acquire()
        self._log.debug("start relation lifecycle")
        try:
            # Start the hook execution scheduler.
            self._scheduler.run()
            # Create a watcher if we don't have one yet.
            if self._watcher is None:
                self._watcher = yield self._unit_relation.watch_related_units(
                    self._scheduler.notify_change)
            # And start the watcher.
            yield self._watcher.start()
        finally:
            self._run_lock.release()

    @inlineCallbacks
    def stop(self):
        """Stop watching changes and stop executing relation change hooks.
        """
        yield self._run_lock.acquire()
        self._log.debug("stop relation lifecycle")
        try:
            self._watcher.stop()
            self._scheduler.stop()
        finally:
            yield self._run_lock.release()

    @inlineCallbacks
    def depart(self):
        """Inform the formula that the service has departed the relation.
        """
        self._log.debug("depart relation lifecycle")
        change = RelationChange(self._relation_name, "departed", "")
        context = self._scheduler.get_hook_context(change)
        hook_name = "%s-relation-broken" % self._relation_name
        yield self._execute_change_hook(context, change, hook_name)
