import logging
import yaml
import csv
import os

from twisted.internet.defer import inlineCallbacks, returnValue

from ensemble.unit.tests.test_lifecycle import LifecycleTestBase
from ensemble.unit.lifecycle import UnitLifecycle, UnitRelationLifecycle
from ensemble.unit.workflow import (
    UnitWorkflow, UnitWorkflowState, RelationWorkflow, RelationWorkflowState)


class WorkflowTestBase(LifecycleTestBase):

    @inlineCallbacks
    def assertState(self, workflow, state):
        workflow_state = yield workflow.get_state()
        self.assertEqual(workflow_state, state)

    @inlineCallbacks
    def read_persistent_state(self, unit=None, history_id=None, workflow=None):
        unit = unit or self.states["unit"]
        history_id = history_id or unit.unit_name
        data, stat = yield self.client.get("/units/%s" % unit.internal_id)
        workflow = workflow or self.workflow

        state = open(workflow.state_file_path).read()
        history = open(workflow.state_history_path)
        zk_state = yaml.load(data)["workflow_state"]
        returnValue((yaml.load(state),
                    [yaml.load(r[0]) for r in csv.reader(history)],
                     yaml.load(zk_state[history_id])))


class UnitWorkflowTest(WorkflowTestBase):

    @inlineCallbacks
    def setUp(self):
        yield super(UnitWorkflowTest, self).setUp()
        yield self.setup_default_test_relation()
        self.lifecycle = UnitLifecycle(
            self.client, self.states["unit"], self.states["service"],
            self.unit_directory, self.executor)

        self.ensemble_directory = self.makeDir()
        self.state_directory = self.makeDir(
            path=os.path.join(self.ensemble_directory, "state"))

        self.workflow = UnitWorkflowState(
            self.client, self.states["unit"], self.lifecycle, UnitWorkflow,
            self.state_directory)

    @inlineCallbacks
    def test_install(self):
        file_path = self.makeFile()
        self.write_hook(
            "install", "#!/bin/bash\necho installed >> %s\n" % file_path)

        result = yield self.workflow.fire_transition("install")

        self.assertTrue(result)
        current_state = yield self.workflow.get_state()
        self.assertEqual(current_state, "installed")

        f_state, history, zk_state = yield self.read_persistent_state()

        self.assertEqual(f_state, zk_state)
        self.assertEqual(f_state,
                         {"state": "installed", "state_variables": {}})
        self.assertEqual(history,
                         [{"state": "installed", "state_variables": {}}])

    @inlineCallbacks
    def test_install_with_error(self):
        """If the install hook fails, the workflow is transition to the
        install_error state.
        """
        result = yield self.workflow.fire_transition("install")
        self.assertFalse(result)
        current_state = yield self.workflow.get_state()
        yield self.assertEqual(current_state, "install_error")
        self.write_hook("install", "#!/bin/bash\necho hello\n")
        result = yield self.workflow.fire_transition("retry_install")
        yield self.assertState(self.workflow, "installed")

    @inlineCallbacks
    def test_start(self):
        file_path = self.makeFile()
        self.write_hook(
            "install", "#!/bin/bash\necho installed >> %s\n" % file_path)
        self.write_hook(
            "start", "#!/bin/bash\necho start >> %s\n" % file_path)
        self.write_hook(
            "stop", "#!/bin/bash\necho stop >> %s\n" % file_path)

        result = yield self.workflow.fire_transition("install")
        self.assertTrue(result)

        result = yield self.workflow.fire_transition("start")
        self.assertTrue(result)

        current_state = yield self.workflow.get_state()
        self.assertEqual(current_state, "started")

        f_state, history, zk_state = yield self.read_persistent_state()

        self.assertEqual(f_state, zk_state)
        self.assertEqual(f_state,
                         {"state": "started", "state_variables": {}})
        self.assertEqual(history,
                         [{"state": "installed", "state_variables": {}},
                          {"state": "started", "state_variables": {}}])

        yield self.lifecycle.stop()

    @inlineCallbacks
    def test_start_with_error(self):
        """Executing the start  transition, results in the workflow going
        to the up state.
        """
        self.write_hook("install", "#!/bin/bash\necho hello\n")
        result = yield self.workflow.fire_transition("install")
        self.assertTrue(result)
        result = yield self.workflow.fire_transition("start")
        self.assertFalse(result)
        current_state = yield self.workflow.get_state()
        self.assertEqual(current_state, "start_error")

        self.write_hook("start", "#!/bin/bash\necho hello\n")
        result = yield self.workflow.fire_transition("retry_start")
        yield self.assertState(self.workflow, "started")
        # If we don't stop, we'll end up with the relation lifecycle
        # watches firing in the background when the test stops.
        self.write_hook("stop", "#!/bin/bash\necho hello\n")
        result = yield self.workflow.fire_transition("stop")
        yield self.assertState(self.workflow, "stopped")

    def test_stop(self):
        """Executing the stop transition, results in the workflow going
        to the down state.
        """
        file_path = self.makeFile()
        self.write_hook(
            "install", "#!/bin/bash\necho installed >> %s\n" % file_path)
        self.write_hook(
            "start", "#!/bin/bash\necho start >> %s\n" % file_path)
        self.write_hook(
            "stop", "#!/bin/bash\necho stop >> %s\n" % file_path)
        result = yield self.workflow.fire_transition("install")
        result = yield self.workflow.fire_transition("start")
        result = yield self.workflow.fire_transition("stop")
        self.assertTrue(result)
        current_state = yield self.workflow.get_state()
        self.assertEqual(current_state, "stopped")
        f_state, history, zk_state = yield self.read_persistent_state()
        self.assertEqual(f_state, zk_state)
        self.assertEqual(f_state,
                         {"state": "stopped", "state_variables": {}})

        self.assertEqual(history,
                         [{"state": "installed", "state_variables": {}},
                          {"state": "started", "state_variables": {}},
                          {"state": "stopped", "state_variables": {}}])

    @inlineCallbacks
    def test_stop_with_error(self):
        self.write_hook("install", "#!/bin/bash\necho hello\n")
        self.write_hook("start", "#!/bin/bash\necho hello\n")
        result = yield self.workflow.fire_transition("install")
        self.assertTrue(result)
        result = yield self.workflow.fire_transition("start")
        self.assertTrue(result)
        result = yield self.workflow.fire_transition("stop")
        self.assertFalse(result)
        current_state = yield self.workflow.get_state()
        self.assertEqual(current_state, "stop_error")
        self.write_hook("stop", "#!/bin/bash\necho hello\n")
        result = yield self.workflow.fire_transition("retry_stop")
        yield self.assertState(self.workflow, "stopped")


class UnitRelationWorkflowTest(WorkflowTestBase):

    @inlineCallbacks
    def setUp(self):
        yield super(UnitRelationWorkflowTest, self).setUp()
        yield self.setup_default_test_relation()
        self.relation_name = self.states["service_relation"].relation_name
        self.ensemble_directory = self.makeDir()
        self.log_stream = self.capture_logging(
            "unit.relation.lifecycle", logging.DEBUG)

        self.lifecycle = UnitRelationLifecycle(
            self.client,
            self.states["unit_relation"],
            self.relation_name,
            self.unit_directory,
            self.executor)

        self.state_directory = self.makeDir(
            path=os.path.join(self.ensemble_directory, "state"))

        self.workflow = RelationWorkflowState(
            self.lifecycle, self.client, RelationWorkflow,
            self.states["unit_relation"],
            self.state_directory)

    @inlineCallbacks
    def test_up_down_cycle(self):
        """The workflow can be transition from up to down, and back.
        """
        self.write_hook("%s-relation-changed" % self.relation_name,
                        "#!/bin/bash\nexit 0\n")

        yield self.workflow.fire_transition("start")
        yield self.assertState(self.workflow, "up")

        hook_executed = self.wait_on_hook("app-relation-changed")

        # Add a new unit, and this will be scheduled by the time
        # we finish stopping.
        yield self.add_opposite_service_unit(self.states)
        yield self.workflow.fire_transition("stop")
        yield self.assertState(self.workflow, "down")
        self.assertFalse(hook_executed.called)

        # Currently if we restart, we will only see the previously
        # queued event, as the last watch active when a lifecycle is
        # stopped, may already be in flight and will be scheduled, and
        # will be executed when the lifecycle is started. However any
        # events that may have occured after the lifecycle is stopped
        # are currently ignored and un-notified.
        yield self.workflow.fire_transition("restart")
        yield self.assertState(self.workflow, "up")
        yield hook_executed

        f_state, history, zk_state = yield self.read_persistent_state(
            history_id=self.workflow.get_zk_history_id())

        self.assertEqual(f_state, zk_state)
        self.assertEqual(f_state,
                         {"state": "up", "state_variables": {}})

        self.assertEqual(history,
                         [{"state": "up", "state_variables": {}},
                          {"state": "down", "state_variables": {}},
                          {"state": "up", "state_variables": {}}])

    @inlineCallbacks
    def test_change_hook_with_error(self):
        """An error while processing a change hook, results
        in the workflow transitioning to the down state.
        """
        self.capture_logging("unit.relation.lifecycle", logging.DEBUG)
        self.write_hook("%s-relation-changed" % self.relation_name,
                        "#!/bin/bash\nexit 1\n")

        current_state = yield self.workflow.get_state()
        self.assertEqual(current_state, None)
        yield self.workflow.fire_transition("start")
        yield self.assertState(self.workflow, "up")
        current_state = yield self.workflow.get_state()

        # Add a new unit, and wait for the broken hook to result in
        # the transition to the down state.
        yield self.add_opposite_service_unit(self.states)
        yield self.wait_on_state(self.workflow, "down")

        f_state, history, zk_state = yield self.read_persistent_state(
            history_id=self.workflow.get_zk_history_id())

        self.assertEqual(f_state, zk_state)
        error = "Error processing '%s': exit code 1." % (
            os.path.join(self.unit_directory,
                         "formula", "hooks", "app-relation-changed"))

        self.assertEqual(f_state,
                         {"state": "down",
                          "state_variables": {
                              "change_type": "joined",
                              "error_message": error}})

    @inlineCallbacks
    def test_depart(self):
        """When the workflow is transition to the down state, a relation
        broken hook is executed, and the unit stops responding to relation
        changes.
        """
        self.write_hook("%s-relation-changed" % self.relation_name,
                        "#!/bin/bash\necho hello\n")
        self.write_hook("%s-relation-broken" % self.relation_name,
                        "#!/bin/bash\necho hello\n")

        results = []

        def collect_executions(*args):
            results.append(args)

        yield self.workflow.fire_transition("start")
        yield self.assertState(self.workflow, "up")

        wait_on_hook = self.wait_on_hook("app-relation-changed")
        states = yield self.add_opposite_service_unit(self.states)
        yield wait_on_hook

        wait_on_hook = self.wait_on_hook("app-relation-broken")
        wait_on_state = self.wait_on_state(self.workflow, "departed")
        yield self.workflow.fire_transition("depart")
        yield wait_on_hook
        yield wait_on_state

        # verify further changes to the related unit, don't result in
        # hook executions.
        self.executor.set_observer(collect_executions)
        states["unit_relation"].set_data(yaml.dump(dict(a=1)))
        yield self.sleep(0.1)
        self.assertFalse(results)

    def test_lifecycle_attribute(self):
        """The workflow lifecycle is accessible from the workflow."""
        self.assertIdentical(self.workflow.lifecycle, self.lifecycle)

    @inlineCallbacks
    def test_depart_hook_error(self):
        """A depart hook error, still results in a transition to the
        departed state with a state variable noting the error."""

        self.write_hook("%s-relation-changed" % self.relation_name,
                        "#!/bin/bash\necho hello\n")
        self.write_hook("%s-relation-broken" % self.relation_name,
                        "#!/bin/bash\nexit 1\n")

        error_output = self.capture_logging("unit.relation.workflow")

        results = []

        def collect_executions(*args):
            results.append(args)

        yield self.workflow.fire_transition("start")
        yield self.assertState(self.workflow, "up")

        wait_on_hook = self.wait_on_hook("app-relation-changed")
        states = yield self.add_opposite_service_unit(self.states)
        yield wait_on_hook

        wait_on_hook = self.wait_on_hook("app-relation-broken")
        wait_on_state = self.wait_on_state(self.workflow, "departed")
        yield self.workflow.fire_transition("depart")
        yield wait_on_hook
        yield wait_on_state

        # verify further changes to the related unit, don't result in
        # hook executions.
        self.executor.set_observer(collect_executions)
        states["unit_relation"].set_data(yaml.dump(dict(a=1)))
        yield self.sleep(0.1)
        self.assertFalse(results)

        # Verify final state and log output.
        msg = "Depart hook error, ignoring: "
        error_msg = "Error processing "
        error_msg += repr(os.path.join(
            self.unit_directory, "formula", "hooks",
            "app-relation-broken"))
        error_msg += ": exit code 1."

        self.assertEqual(
            error_output.getvalue(), (msg + error_msg + "\n"))
        current_state = yield self.workflow.get_state()
        self.assertEqual(current_state, "departed")

        f_state, history, zk_state = yield self.read_persistent_state(
            history_id=self.workflow.get_zk_history_id())

        self.assertEqual(f_state, zk_state)
        self.assertEqual(f_state,
                         {"state": "departed",
                          "state_variables": {
                              "change_type": "depart",
                              "error_message": error_msg}})

    def test_depart_down(self):
        """When the workflow is transition to the down state, a relation
        broken hook is executed, and the unit stops responding to relation
        changes.
        """
        self.write_hook("%s-relation-changed" % self.relation_name,
                        "#!/bin/bash\necho hello\n")
        self.write_hook("%s-relation-broken" % self.relation_name,
                        "#!/bin/bash\necho hello\n")

        results = []

        def collect_executions(*args):
            results.append(args)

        yield self.workflow.fire_transition("start")
        yield self.assertState(self.workflow, "up")

        yield self.workflow.fire_transition("stop")
        yield self.assertState(self.workflow, "down")

        states = yield self.add_opposite_service_unit(self.states)

        wait_on_hook = self.wait_on_hook("app-relation-broken")
        wait_on_state = self.wait_on_state(self.workflow, "departed")

        yield self.workflow.fire_transition("depart")
        yield wait_on_hook
        yield wait_on_state

        # verify further changes to the related unit, don't result in
        # hook executions.
        self.executor.set_observer(collect_executions)
        states["unit_relation"].set_data(yaml.dump(dict(a=1)))
        yield self.sleep(0.1)
        self.assertFalse(results)
