import yaml
import csv
import os
import logging

from twisted.internet.defer import inlineCallbacks, returnValue

from txzookeeper.utils import retry_change

from ensemble.errors import FormulaInvocationError, FormulaError, FileNotFound
from ensemble.lib.statemachine import (
    WorkflowState, Workflow, Transition, TransitionError)


UnitWorkflow = Workflow(
    Transition("install", "Install", None, "installed",
               error_transition_id="error_install"),
    Transition("error_install", "Install Error", None, "install_error"),
    Transition("retry_install", "Retry Install", "install_error", "installed"),
    Transition("start", "Start", "installed", "started",
               error_transition_id="error_start"),
    Transition("error_start", "Start Error", "installed", "start_error"),
    Transition("retry_start", "Retry Start", "start_error", "started"),
    Transition("stop", "Stop", "started", "stopped",
               error_transition_id="error_stop"),
    Transition("error_stop", "Stop Error", "started", "stop_error"),
    Transition("retry_stop", "Retry Stop", "stop_error", "stopped"),
    )


# There's been some discussion, if we should have per change type error states
# here, corresponding to the different changes that the relation-changed hook
# is invoked for. The important aspects to capture are both observability of
# error type locally and globally (zk), and per error type and instance
# recovery of the same. To provide for this functionality without additional
# states, the error information (change type, and error message) are captured
# in state variables which are locally and globally observable. Future
# extension of the restart transition action, will allow for customized
# recovery based on the change type state variable. Effectively this
# differs from the unit definition, in that it collapses three possible error
# states, into a behavior off switch. A separate state will be needed to
# denote departing.

RelationWorkflow = Workflow(
    Transition("start", "Start", None, "up"),
    Transition("stop", "Stop", "up", "down"),
    Transition("restart", "Restart", "down", "up"),
    Transition("depart", "Relation broken", "up", "departed"),
    Transition("down_depart", "Relation broken", "down", "departed"),
    )


class PersistentWorkflowStateBase(object):

    # Used to construct a state_file_path with the state directory.
    state_file_path_suffix = "state.txt"

    # Used to construct a state_history_path with the state directory.
    state_history_path_suffix = "history.txt"

    def _write_state_file(self, workflow_state):
        with open(self.state_file_path, "w") as handle:
            handle.write(workflow_state)
            handle.flush()

    def _read_state_file(self):
        if not os.path.exists(self.state_file_path):
            return None

        with open(self.state_file_path, "r") as handle:
            return handle.read()

    def _append_history(self, workflow_state):
        """Append workflow state to history file.

        Internally the history file is stored a csv, with a new
        row per entry. CSV escaping is used for escaping, new
        lines.
        """
        with open(self.state_history_path, "a") as handle:
            writer = csv.writer(handle)
            writer.writerow((workflow_state,))
            handle.flush()

    def get_zk_history_id(self):
        """Get the id of this workflow within the zookeeper stored history.
        """

    def get_zk_history_path(self):
        """Get th path to the zookeeper stored workflow state.
        """

    def _write_state_zk(self, workflow_state):

        state_id = self.get_zk_history_id()
        history_path = self.get_zk_history_path()

        def update_state(content, stat):
            unit_data = yaml.load(content)
            if not unit_data:
                unit_data = {}

            persistent_workflow = unit_data.setdefault("workflow_state", {})
            persistent_workflow[state_id] = workflow_state
            return yaml.dump(unit_data)

        return retry_change(self._client, history_path, update_state)


class UnitWorkflowState(WorkflowState, PersistentWorkflowStateBase):

    def __init__(self, client, unit_state, lifecycle, workflow,
                 state_directory):

        super(UnitWorkflowState, self).__init__(workflow)
        self._client = client
        self._lifecycle = lifecycle
        self._unit_state = unit_state
        self._zk_unit_path = "/units/%s" % unit_state.internal_id
        self._state_directory = state_directory

        self.state_file_path = "%s/%s-%s" % (
            self._state_directory,
            unit_state.unit_name.replace("/", "-"),
            self.state_file_path_suffix)

        self.state_history_path = "%s/%s-%s" % (
            self._state_directory,
            unit_state.unit_name.replace("/", "-"),
            self.state_history_path_suffix)

    def get_zk_history_id(self):
        """fetch the id of this workflow within the zookeeper stored history.
        """
        return self._unit_state.unit_name

    def get_zk_history_path(self):
        return self._zk_unit_path

    def _load(self):
        """Retrieve the workflow state.
        """
        content = self._read_state_file()
        if content is None:
            return {"state": None}
        return yaml.load(content)

    def _store(self, state_dict):
        """Persist the workflow state.
        """
        state_serialized = yaml.dump(state_dict)
        self._write_state_file(state_serialized)
        self._append_history(state_serialized)
        return self._write_state_zk(state_serialized)

    @inlineCallbacks
    def _invoke_lifecycle(self, method):
        try:
            result = yield method()
        except (FileNotFound, FormulaError, FormulaInvocationError), e:
            raise TransitionError(e)
        returnValue(result)

    # Transition Actions
    def do_install(self):
        return self._invoke_lifecycle(self._lifecycle.install)

    def do_start(self):
        return self._invoke_lifecycle(self._lifecycle.start)

    def do_stop(self):
        return self._invoke_lifecycle(self._lifecycle.stop)

    def do_retry_start(self):
        return self._invoke_lifecycle(self._lifecycle.start)

    def do_retry_stop(self):
        self._invoke_lifecycle(self._lifecycle.stop)

    def do_retry_install(self):
        return self._invoke_lifecycle(self._lifecycle.install)


class RelationWorkflowState(WorkflowState, PersistentWorkflowStateBase):

    def __init__(self, lifecycle, client, workflow, unit_relation,
                 state_directory):
        super(RelationWorkflowState, self).__init__(workflow)
        self._lifecycle = lifecycle
        # Catch any related-change hook errors
        self._lifecycle.set_hook_error_handler(self.on_hook_error)
        self._client = client
        self._unit_relation = unit_relation
        self._zk_unit_path = "/units/%s" % (unit_relation.internal_unit_id)
        self._workflow_state_directory = state_directory
        self._log = logging.getLogger("unit.relation.workflow")

        self.state_file_path = "%s/%s-%s-%s" % (
            self._workflow_state_directory,
            unit_relation.internal_unit_id,
            unit_relation.internal_relation_id,
            self.state_file_path_suffix)

        self.state_history_path = "%s/%s-%s-%s" % (
            self._workflow_state_directory,
            unit_relation.internal_unit_id,
            unit_relation.internal_relation_id,
            self.state_history_path_suffix)

    @property
    def lifecycle(self):
        return self._lifecycle

    def get_zk_history_id(self):
        return self._unit_relation.internal_relation_id

    def get_zk_history_path(self):
        return self._zk_unit_path

    def _load(self):
        content = self._read_state_file()
        if content is None:
            return {"state": None}
        return yaml.load(content)

    def _store(self, state_dict):
        """Persist the workflow state.
        """
        state_serialized = yaml.dump(state_dict)
        self._write_state_file(state_serialized)
        self._append_history(state_serialized)
        return self._write_state_zk(state_serialized)

    @inlineCallbacks
    def on_hook_error(self, relation_change, error):
        """Handle relation-change hook errors.

        Invoked by the hook scheduler on error. The relation-change
        hooks are executed out of band, as a result of watch
        invocations. We have the relation lifecycle accept this method
        as an error handler, so we can drive workflow changes as a
        result of hook errors.

        @param: relation_change: describes the change for which the
        hook is being invoked.

        @param: error: The error from hook invocation.
        """
        yield self.fire_transition("stop",
                                   change_type=relation_change.change_type,
                                   error_message=str(error))

    @inlineCallbacks
    def do_stop(self):
        """Transition the workflow to the 'down' state.

        Turns off the unit-relation lifecycle monitoring and hook execution.
        """
        yield self._lifecycle.stop()

    @inlineCallbacks
    def do_restart(self):
        """Transition the workflow to the 'up' state.

        Turns on the unit-relation lifecycle monitoring and hook execution.
        """
        yield self._lifecycle.start()

    @inlineCallbacks
    def do_start(self):
        """Transition the workflow to the 'up' state.

        Turns on the unit-relation lifecycle monitoring and hook execution.
        """
        yield self._lifecycle.start()

    @inlineCallbacks
    def do_depart(self):
        """Transition a relation to the departed state, from the up state.
        """
        # Stop related unit watches and change hook execution.
        yield self._lifecycle.stop()
        result = yield self._do_depart()
        returnValue(result)

    def do_down_depart(self):
        """Transition a relation to the departed state, from the down state.
        """
        return self._do_depart()

    @inlineCallbacks
    def _do_depart(self):
        """Execute the depart hook.

        We ignore hook errors, as we won't logically process any additional
        events for the relation once it doesn't exist. However we do
        note the error in the log.
        """
        # To avoid the relation-changed hook error handler being used,
        # set the handler to None, so the exception is raised.
        self._lifecycle.set_hook_error_handler(None)

        try:
            yield self._lifecycle.depart()
        except Exception, e:
            self._log.error("Depart hook error, ignoring: %s", str(e))
            returnValue({"change_type": "depart",
                         "error_message": str(e)})
